#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that repairs a graph based on a reference.
"""
import networkx as nx

from ..molecule import Block
from .processor import Processor
from ..graph_utils import *  # FIXME  # pylint: disable=unused-wildcard-import
from ..ismags import ISMAGS
from ..log_helpers import StyleAdapter, get_logger
from ..utils import format_atom_string, are_all_equal

LOGGER = StyleAdapter(get_logger(__name__))


def get_default(dictionary, attr, default):
    """
    Functions like :meth:`dict.get`, except that when `attr` is in `dictionary`
    and `dictionary[attr]` is `None`, it will return `default`.

    Parameters
    ----------
    dictionary: dict
    attr: collections.abc.Hashable
    default

    Returns
    -------
    object
        The value of `dictionary[attr]` if `attr` is in `dictionary` and
        `dictionary[attr]` is not None. `default otherwise.`
    """
    item = dictionary.get(attr, None)
    if item is None:
        item = default
    return item


def _node_equal(node1, node2):
    return node1.get('atomname') == node2.get('atomname')


def _patch_modification(block, modification):
    """
    Applies, in order, modifications to block and returns a new Block.
    Modifications are applied by overlaying the anchor of a modification with
    the block and adding the remaining nodes and edges.
    """
    anchor_idxs = set()
    for mod_idx in modification:
        if not modification.nodes[mod_idx]['PTM_atom']:
            anchor_idxs.add(mod_idx)
    anchor = nx.subgraph(modification, anchor_idxs)
    non_anchor_idxs = set(modification) - anchor_idxs
    non_anchor = nx.subgraph(modification, non_anchor_idxs)

    ismags = ISMAGS(block, anchor, node_match=_node_equal)
    anchor_block_to_mod = list(ismags.subgraph_isomorphisms_iter())
    if not anchor_block_to_mod:
        LOGGER.error("Modification {} doesn't fit on Block {}", modification.name,
                     block.name)
        raise ValueError("Cannot apply modification to block")
    # This probably can't happen, since atoms in Modifications /should/ have
    # atomnames. Which should be unique. So, nocover.
    elif len(anchor_block_to_mod) > 1:  # pragma: nocover
        LOGGER.error("Modification {} fits on Block {} in {} ways",
                     modification.name, block.name, len(anchor_block_to_mod))  # pragma: nocover
        raise ValueError("Cannot apply modification to block")  # pragma: nocover

    anchor_block_to_mod = anchor_block_to_mod[0]
    anchor_mod_to_block = {val: key for key, val in anchor_block_to_mod.items()}

    result = Block(nx.union(block, non_anchor, rename=(None, modification.name+'-')))
    # Mark the modified atoms with the specified modification, so canonicalize
    # modifications has an easier job
    for idx in anchor_block_to_mod:
        node_mods = result.nodes[idx].get('modifications', [])
        if modification not in node_mods:
            result.nodes[idx]['modifications'] = node_mods + [modification]
    for mod_idx in non_anchor_idxs:
        idx = '{}-{}'.format(modification.name, mod_idx)
        node_mods = result.nodes[idx].get('modifications', [])
        if modification not in node_mods:
            result.nodes[idx]['modifications'] = node_mods + [modification]
    for mod_idx, mod_jdx in modification.edges_between(anchor_idxs, non_anchor_idxs):
        if mod_idx in non_anchor_idxs:
            mod_idx, mod_jdx = mod_jdx, mod_idx
        # mod_idx is always in anchor, and thus in block.
        block_idx = anchor_mod_to_block[mod_idx]
        # Because of how nx.union renamed nodes in non_anchor
        mod_jdx = '{}-{}'.format(modification.name, mod_jdx)
        result.add_edge(block_idx, mod_jdx)
    return result


def _get_reference_residue(residue, force_field):
    """
    Uses the 'mutation', 'modify' and 'resname' attributes of `residue` to find
    or generate the correct reference residue, based on `force_field`. Mutation
    takes priority over resname, and the corresponding block will be taken from
    the FF. Afterwards, all modifications in modify will be taken from the FF
    and applied.

    Returns the generated reference block.
    """
    if 'mutation' in residue:
        mutation = residue['mutation']
        if not are_all_equal(mutation):
            message = 'Can only mutate residue {}-{}{} once, {} mutations were requested.'
            LOGGER.error(message, residue['chain'], residue['resname'],
                         residue['resid'], len(mutation))
            raise ValueError(message.format(residue['chain'], residue['resname'],
                                            residue['resid'], len(mutation)))
        mutation = mutation[0]
        LOGGER.info('Mutating residue {}-{}{} to {}',
                    residue['chain'], residue['resname'], residue['resid'],
                    mutation)
        resname = mutation
    else:
        resname = residue['resname']
    reference_block = force_field.reference_graphs[resname]

    if 'modification' in residue:
        modifications = residue['modification']
        for mod_name in modifications:
            LOGGER.info('Applying modification {} to residue {}-{}{}',
                        mod_name, residue['chain'], resname, residue['resid'])
            if mod_name != 'none':
                mod = force_field.modifications[mod_name]
                reference_block = _patch_modification(reference_block, mod)
        for node_idx in reference_block:
            reference_block.nodes[node_idx]['modification'] = modifications
    if 'mutation' in residue:
        for node_idx in reference_block:
            reference_block.nodes[node_idx]['mutation'] = mutation

    return reference_block


def make_reference(mol):
    """
    Takes an molecule graph (e.g. as read from a PDB file), and finds and
    returns the graph how it should look like, including all matching nodes
    between the input graph and the references.
    Requires residuenames to be correct.

    Notes
    -----
        The match between hydrogren atoms need not be perfect. See the
        documentation of ``isomorphism``.

    Parameters
    ----------
    mol : networkx.Graph
        The graph read from e.g. a PDB file. Required node attributes:

        :resname: The residue name.
        :resid: The residue id.
        :chain: The chain identifier.
        :element: The element.
        :atomname: The atomname.

    Returns
    -------
    networkx.Graph
        The constructed reference graph with the following node attributes:

        :resid: The residue id.
        :resname: The residue name.
        :chain: The chain identifier.
        :found: The residue subgraph from the PDB file.
        :reference: The residue subgraph used as reference.
        :match: A dictionary describing how the reference corresponds
            with the provided graph. Keys are node indices of the
            reference, values are node indices of the provided graph.
    """
    reference_graph = nx.Graph()
    residues = make_residue_graph(mol)
    symmetry_cache = {}
    LOGGER.debug('Making reference graph', type='step')
    for residx in residues:
        # TODO: make separate function for just one residue.
        # TODO: Merge degree 1 nodes (hydrogens!) with the parent node. And
        # check whether the node degrees match?

        resname = residues.nodes[residx]['resname']
        resid = residues.nodes[residx]['resid']
        chain = residues.nodes[residx]['chain']
        residue = residues.nodes[residx]['graph']
        reference = _get_reference_residue(residues.nodes[residx], mol.force_field)
        if 'mutation' in residues.nodes[residx]:
            # We need to do this, since we need to propagate all common
            # attributes in the input residue to any reconstructed atoms.
            # make_residue_graph already picks up on all the common attributes
            # and sets those as residue/node attributes, but any mutation needs
            # to be propagated as well, otherwise the newly constructed atoms
            # will end up with the old residue name. This way, we can just add
            # *all* attributes in the residue graph to the new atoms.
            resname = residues.nodes[residx]['mutation'][0]
            residues.nodes[residx]['resname'] = resname
        add_element_attr(reference)
        add_element_attr(residue)
        # We are going to sort the nodes of reference and residue by atomname.
        # We do this, because the ISMAGS algorithm prefers to match nodes with
        # lower IDs.
        # Get a \uFFFF for every node that doesn't have an atomname attribute
        # or when it's None, since that sorts higher than letters, giving them
        # the lowest priority in ISMAGS.

        res_names = {idx: get_default(residue.nodes[idx], 'atomname', '\uFFFF') for idx in residue}
        ref_names = {idx: get_default(reference.nodes[idx], 'atomname', '\uFFFF') for idx in reference}

        # Sort the nodes such that any atomnames that are common to both
        # reference and residue are first, and then the rest.
        # Also, sort it all by atomname. This is combined in one by sorting by
        # the tuple (not common, atomname). False < True.

        # If we want to relabel the nodes in-place we need to find new
        # non-overlapping labels. The easiest way of doing this is by turning
        # them into tuples. But this makes everything slow; probably because
        # ISMAGS does quite a lot of inequality comparisons, and those are way
        # faster for str/int. So, sacrifice the memory, and relabel by making a
        # new copy.

        # TODO: include a geometric alignment in the sorting. Humans are really
        #       good at solving isomorphism problems iff graphs look alike. We
        #       can do a similar trick here by rot+trans aligning the given
        #       residue with a reference conformation. And then sort by
        #       distance
        new_residue_names = {old: new for new, old in enumerate(sorted(
            residue,
            key=lambda jdx: (res_names[jdx] not in ref_names.values(), res_names[jdx])  # pylint: disable=cell-var-from-loop
        ))}
        new_reference_names = {old: new for new, old in enumerate(sorted(
            reference,
            key=lambda jdx: (ref_names[jdx] not in res_names.values(), ref_names[jdx])  # pylint: disable=cell-var-from-loop
        ))}

        old_res_names = {v: k for k, v in new_residue_names.items()}
        old_ref_names = {v: k for k, v in new_reference_names.items()}

        # It would be nice if we were able to relabel them in-place, but it
        # seems to make everything slower. See above.
        res_copy = nx.relabel_nodes(residue, new_residue_names, copy=True)
        ref_copy = nx.relabel_nodes(reference, new_reference_names, copy=True)

        LOGGER.debug('Matching residue {}-{}{} to its reference', chain, resname, resid, type='step')

        # If we assume residue > reference the tests run *way* faster, but the
        # actual program becomes *much* *much* slower.
        # TODO: swap ref_copy and res_copy so that the smaller graph is the
        #       "subgraph"?
        ismags = ISMAGS(ref_copy, res_copy,
                        node_match=nx.isomorphism.categorical_node_match('element', None),
                        cache=symmetry_cache)
        # Finding the largest common subgraph is expensive, but the first step
        # is to try and find a subgraph isomorphism between
        # residue <= reference, so best case it makes no difference, and worst
        # case we avoid trying to find that isomorphism twice.
        match_iter = ismags.largest_common_subgraph()
        try:
            # We take only the first found match, since because the nodes are
            # sorted by atomname, and ISMAGS prefers to take nodes with low ID,
            # that match should have most matching atomnames.
            match = next(match_iter)
        except StopIteration:
            LOGGER.error("Can't find isomorphism between {}-{}{} and its "
                         "reference.", chain, resname, resid, type='inconsistent-data')
            continue
        # TODO: Since we only have one isomorphism we don't know whether the
        # assigment we're making is ambiguous. So iff the residue is small
        # enough (or a flag is set, whatever), also find the second isomorphism
        # and check whether it has the same number of correct atomnames. If so,
        # issue a warning and carry on. We can't do this for all residues,
        # since that takes a cup of coffee.

        # "unsort" the matches
        match = {old_ref_names[ref]: old_res_names[res] for ref, res in match.items()}

        # The residue graph has attributes which are common to all atoms in the
        # input residue, so propagate those to new atoms. These attributes are
        # things like residue name (see above in case of mutations), resid,
        # insertion code, etc.
        reference_graph.add_node(residx, reference=reference, found=residue,
                                 match=match, **residues.nodes[residx])
    reference_graph.add_edges_from(residues.edges())
    return reference_graph


def repair_residue(molecule, ref_residue, include_graph):
    """
    Rebuild missing atoms and canonicalize atomnames
    """
    # Rebuild missing atoms and canonicalize atomnames
    missing = []
    # Step 1: find all missing atoms. Canonicalize names while we're at it.
    reference = ref_residue['reference']
    found = ref_residue['found']
    match = ref_residue['match']

    resid = ref_residue['resid']
    resname = ref_residue['resname']
    LOGGER.debug('Repairing residue {}{}', resname, resid, type='step')
    for ref_idx in reference:
        if ref_idx in match:
            res_idx = match[ref_idx]
            node = molecule.nodes[res_idx]
            if include_graph:
                node['graph'] = molecule.subgraph([res_idx])
            node.update(reference.nodes[ref_idx])
            # Update found as well to keep found and molecule in line. It would
            # be better to try and figure why found is not a reference, but meh
            found.nodes[res_idx].update(reference.nodes[ref_idx])
        else:
            message = 'Missing atom {}{}:{}'
            args = (resname, resid, reference.nodes[ref_idx]['atomname'])
            if reference.nodes[ref_idx]['element'] != 'H':
                LOGGER.info(message, *args, type='missing-atom')
            else:
                # These are logged *below* debug level. Otherwise your screen
                # fills up pretty fast.
                LOGGER.log(5, message, *args, type='missing-atom')
            missing.append(ref_idx)
    # Step 2: try to add all missing atoms one by one. As long as we added
    # *something* the situation changed, and we might be able to place another.
    # We can only place atoms for which we know a neighbour.
    added = True
    while missing and added:
        added = False
        for ref_idx in missing:
            # See if the atom we want to add has a known neighbour. Otherwise,
            # continue to the next.
            if all(ref_neighbour in missing for ref_neighbour in reference[ref_idx]):
                continue
            added = True
            missing.pop(missing.index(ref_idx))
            # We don't find the lowest available number since that's just
            # asking for problems where you find an atom you don't expect
            # because the old one you were looking for was removed, and it's
            # number was reassigned.
            res_idx = max(molecule) + 1

            # Create the new node
            node = {}
            for key, val in ref_residue.items():
                # Some attributes are only relevant on a residue level, not on
                # an atom level.
                if key not in ('match', 'found', 'reference'):
                    node[key] = val
            node.update(reference.nodes[ref_idx])
            node['atomid'] = res_idx + 1

            match[ref_idx] = res_idx
            molecule.add_node(res_idx, **node)
            found.add_node(res_idx, **node)

            message = "Adding {}"
            args = format_atom_string(node)
            if node['element'] != 'H':
                LOGGER.debug(message, args, type='missing-atom')
            else:
                # These are logged *below* debug level. Otherwise your screen
                # fills up pretty fast.
                LOGGER.log(5, message, args, type='missing-atom')

            neighbours = 0
            for neighbour_ref_idx in reference[ref_idx]:
                try:
                    neighbour_res_idx = match[neighbour_ref_idx]
                except KeyError:
                    continue
                if not molecule.has_edge(neighbour_res_idx, res_idx):
                    molecule.add_edge(neighbour_res_idx, res_idx)
                    neighbours += 1
            assert neighbours != 0

    for ref_idx in missing:
        # TODO: use utils.format_atom_string
        LOGGER.error('Could not reconstruct atom {}{}:{}',
                     reference.nodes[ref_idx]['resname'],
                     resid,
                     reference.nodes[ref_idx]['atomname'],
                     type='missing-atom')


def repair_graph(molecule, reference_graph, include_graph=True):
    """
    Repairs a molecule graph produced based on the information in
    ``reference_graph``. Missing atoms will be added and atom- and residue-
    names will be canonicalized. Atoms not present in ``reference_graph`` will
    have the attribute ``PTM_atom`` set to ``True``.

    `molecule` is modified in place. Missing atoms (as per `reference_graph`)
    are added, atom and residue names are canonicalized, and PTM atoms are
    marked.

    If `include_graph` is `True`, then the subgraph corresponding to each node
    is included in the node under the "graph" attribute.

    Parameters
    ----------
    molecule : molecule.Molecule
        The graph read from e.g. a PDB file. Required node attributes:

        :resname: The residue name.
        :resid: The residue id.
        :element: The element.
        :atomname: The atomname.

    reference_graph : networkx.Graph
        The reference graph as produced by ``make_reference``. Required node
        attributes:

        :resid: The residue id.
        :resname: The residue name.
        :found: The residue subgraph from the PDB file.
        :reference: The residue subgraph used as reference.
        :match: A dictionary describing how the reference corresponds
            with the provided graph. Keys are node indices of the
            reference, values are node indices of the provided graph.

    include_graph: bool
        Include the subgraph in the nodes.
    """
    for residx in reference_graph:
        residue = reference_graph.nodes[residx]
        repair_residue(molecule, residue, include_graph=include_graph)
        # Atomnames are canonized, and missing atoms added
        found = reference_graph.nodes[residx]['found']
        match = reference_graph.nodes[residx]['match']

        # Find the PTMs (or termini, or other additions) for *this* residue
        # `extra` is a set of the indices of the nodes from  `found` that have
        # no match in the reference graph.
        # `atachments` is a set of the nodes from `found` that have a match in
        # the reference and are connected to a node from `extra`.
        # We just stick a label on them for now, these are used by the PTM
        # processor.
        extra = set(found.nodes) - set(match.values())
        for idx in extra:
            molecule.nodes[idx]['PTM_atom'] = True
            found.nodes[idx]['PTM_atom'] = True
            if molecule.nodes[idx].get('mutation') or molecule.nodes[idx].get('modification'):
                molecule.remove_node(idx)

    return molecule


class RepairGraph(Processor):
    def __init__(self, delete_unknown=False, include_graph=True):
        super().__init__()
        self.delete_unknown = delete_unknown
        self.include_graph = include_graph

    def run_molecule(self, molecule):
        molecule = molecule.copy()
        reference_graph = make_reference(molecule)
        repair_graph(molecule, reference_graph, include_graph=self.include_graph)
        return molecule

    def run_system(self, system):
        mols = []
        for idx, molecule in enumerate(system.molecules):
            try:
                new_molecule = self.run_molecule(molecule)
            except KeyError as err:
                if not self.delete_unknown:
                    raise err
                else:
                    LOGGER.warning("Cannot recognize residue {} in  molecule {}. "
                                   "Deleting the molecule.",
                                   str(err), idx, type='unknown-residue')
            else:
                mols.append(new_molecule)
        system.molecules = mols



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the mapping file parser and its utilities.
"""
import textwrap

import networkx as nx
import pytest

from vermouth.forcefield import ForceField
from vermouth.molecule import Block, Link
from vermouth.map_parser import MappingDirector, MappingBuilder
from vermouth.tests.helper_functions import equal_graphs


# PyLint does *not* like pytest fixtures...
# pylint: disable=redefined-outer-name

@pytest.fixture
def force_fields():
    """
    Creates a simple forcefield for tests
    """
    ffa = ForceField(name='test_A')
    block_a = Block(name='block_A', force_field=ffa)
    block_a.add_nodes_from(['A1', 'A2', 'B1', 'B2'])
    for node in block_a.nodes:
        block_a.nodes[node]['atomname'] = node
        block_a.nodes[node]['resname'] = 'block_A'
    block_a.add_edges_from([('A1', 'A2'), ('B1', 'B2')])
    ffa.blocks['block_A'] = block_a
    ffa.modifications['block_A'] = Link(block_a)
    block_b = Block(name='block_B', force_field=ffa)
    block_b.add_node('A', atomname='A', resname='block_B', replace={'atomname': 'B'})
    ffa.blocks['block_B'] = block_b
    return {'test_A': ffa}


def equal_graph_data(found_nodes, found_edges, expected_nodes, expected_edges):
    """
    Returns True iff the data provided produces two equivalent graphs.
    """
    found_graph = nx.Graph()
    found_graph.add_nodes_from(found_nodes)
    found_graph.add_edges_from(found_edges)
    exp_graph = nx.Graph()
    exp_graph.add_nodes_from(expected_nodes)
    exp_graph.add_edges_from(expected_edges)
    return equal_graphs(found_graph, exp_graph, node_attrs=None, edge_attrs=None)


@pytest.fixture
def director(force_fields):
    """
    Produces a MappinDirector for parsing.
    """
    return MappingDirector(force_fields=force_fields, builder=MappingBuilder())


@pytest.mark.parametrize('lines, expected', (
    (  # Simple case
        """
        [modification]
        [ to ]
        martini22

        [ from ]
        universal
        """,
        {
            'ff_from': 'universal',
            'ff_to': 'martini22'
        }
    ),
    (  # Test space significance on section headers
        """
        [block ]
        [from ]
        martini30
        [ to]
        universal
        """,
        {
            'ff_from': 'martini30',
            'ff_to': 'universal'
        }
    ),
    (  # Test comments on values
        """
        [block ]
        [from ]
        martini30; comment
        [ to]
        universal ; more comment
        """,
        {
            'ff_from': 'martini30',
            'ff_to': 'universal'
        }
    ),
    (  # Also test comments on section headers
        """
        [block ] ; more comment
        [from ]; comment
        martini30
        [ to]       ; even more comments
        universal
        """,
        {
            'ff_from': 'martini30',
            'ff_to': 'universal'
        }
    ),
    (  # Test simplest to nodes
        """
        [block]
        [ to blocks]
        !B {}
        [ to nodes]
        B:B {"boing": "a"}
        """,
        {
            'block_to': ([(0, {'atomname': 'B', "boing": 'a'})], []),
        }
    ),
    (  # Test to nodes with multiple nodes and block attributes
        """
        [block]
        [ to blocks]
        !B {"plop": 42}
        [ to nodes]
        B:B {"boing": "a"}
        C {}
        """,
        {
            'block_to': ([(0, {'atomname': 'B', "boing": 'a', "plop": 42}),
                          (1, {'atomname': 'C', 'plop': 42})],
                         []),
        }
    ),
    (  # to nodes with identical atomnames
        """
        [block]
        [ to blocks]
        !B {}
        [ to nodes]
        B {"boing": "a"}
        B {}
        """,
        {
            'block_to': ([(0, {'atomname': 'B', "boing": 'a'}),
                          (1, {'atomname': 'B'})],
                         []),
        }
    ),
    (  # to edges, with edge attribute
        """
        [block]
        [ to blocks]
        !B {}
        [ to nodes]
        B {"boing": "a"}
        C
        [ to edges ]
        B:C B:B {"peanut": "butter"}
        """,
        {
            'block_to': ([(0, {'atomname': 'B', "boing": 'a'}),
                          (1, {'atomname': 'C'})],
                         [(0, 1, {"peanut": "butter"})]),
        }
    ),
    (  # Several block IDs for nodes with identical atomnames
        """
        [block]
        [ to blocks]
        !B1 {"resid": 1}
        !B2 {"resid": 2}
        [ to nodes]
        B1:B {"boing": "a"}
        B2:B {}
        [ to edges ]
        B1:B B2:B
        """,
        {
            'block_to': ([(0, {'atomname': 'B', "boing": 'a', 'resid': 1}),
                          (1, {'atomname': 'B', "resid": 2})],
                         [(0, 1, {})]),
        }
    ),
    (  # Test nodes from and to, and mapping
        """
        [modification]
        [ from blocks ]
        !A1 {"ping": false}
        !A2 {"resid": 23}
        [ from nodes ]
        A1:A {"plop": 1}
        A2:A {}
        [ from edges ]
        A1:A A2:A
        [ to blocks]
        !B {}
        [ to nodes]
        B:B {"boing": "a"}
        [ mapping ] ; Mapping section is needed since unmapped from nodes get removed
        A1:A B
        A2:A B
        """,
        {
            'block_from': ([(0, {'atomname': 'A', "plop": 1, 'ping': False}),
                            (1, {'atomname': 'A', 'resid': 23})], [(0, 1, {})]),
            'block_to': ([(0, {'atomname': 'B', "boing": 'a'})], []),
        }
    ),
    (  # Identical to and from block IDs
        """
        [block]
        [ from blocks ]
        !A1 {"ping": false}
        !A2 {"resid": 23}
        [ from nodes ]
        A1:A {"plop": 1}
        A2:A {}
        [ from edges ]
        A1:A A2:A
        [ to blocks]
        !A1 {}
        [ to nodes]
        B {"boing": "a"}
        [ mapping ]
        A1:A B
        A2:A B
        """,
        {
            'block_from': ([(0, {'atomname': 'A', "plop": 1, 'ping': False}),
                            (1, {'atomname': 'A', 'resid': 23})], [(0, 1, {})]),
            'block_to': ([(0, {'atomname': 'B', "boing": 'a'})], []),
            'mapping': {0: {0: 1}, 1: {0: 1}},
            'references': {}
        }

    ),
    (  # Reference atoms and mapping weights
        """
        [modification]
        [ to ]
        ff_a
        [ from]
        ff_b
        [ from blocks ]
        !A1 {"ping": false}
        !A2 {"resid": 23}
        [ from nodes ]
        A1:A {"plop": 1}
        A2:A {}
        [ from edges ]
        A1:A A2:A
        [ to blocks]
        !B {}
        [ to nodes]
        B:B {"boing": "a"}
        [ mapping ]
        A1:A B
        A2:A B 2
        [ reference atoms ]
        B A2:A
        """,
        {
            'block_from': ([(0, {'atomname': 'A', "plop": 1, 'ping': False}),
                            (1, {'atomname': 'A', 'resid': 23})], [(0, 1, {})]),
            'block_to': ([(0, {'atomname': 'B', "boing": 'a'})], []),
            'ff_to': 'ff_a',
            'ff_from': 'ff_b',
            'references': {0: 1},
            'mapping': {0: {0: 1}, 1: {0: 2}}
        }
    ),
    (  # Block shorthand, one node mapping to multiple others
        """
        [ block ]
        [ from blocks ]
        !ALA#1 !GLY#2
        [ to blocks ]
        !A#5 !B13 !C1#3
        [ from nodes ]
        ALA#1:A
        B
        GLY#2:A
        [ to nodes ]
        A#5:A
        B13:A
        B13:B
        C1#3:A
        [ mapping ]
        ALA#1:A A#5:A
        ALA#1:A B13:A
        B       B13:B
        GLY#2:A C1#3:A
        """,
        {
            'block_from': ([(0, {"resname": 'ALA', 'resid': 1, 'atomname': 'A'}),
                            (1, {"resname": 'ALA', 'resid': 1, 'atomname': 'B'}),
                            (2, {"resname": 'GLY', 'resid': 2, 'atomname': 'A'})],
                           []),
            'block_to': ([(0, {"resname": 'A', 'resid': 5, 'atomname': 'A'}),
                          (1, {"resname": 'B13', 'resid': 1, 'atomname': 'A'}),
                          (2, {"resname": 'B13', 'resid': 1, 'atomname': 'B'}),
                          (3, {"resname": 'C1', 'resid': 3, 'atomname': 'A'})],
                         []),
            'mapping': {0: {0: 1, 1: 1}, 1: {2: 1}, 2: {3: 1}}
        }
    ),
    (  # Block shorthand, remove unmapped nodes
        """
        [ block ]
        [ from blocks ]
        !ALA#1 !GLY#2
        [ to blocks ]
        !A#5 !B13 !C1#3
        [ from nodes ]
        ALA#1:A
        B
        GLY#2:A
        [ to nodes ]
        A#5:A
        B13:A
        B13:B
        C1#3:A
        [ mapping ]
        ALA#1:A A#5:A
        ALA#1:A B13:A
        GLY#2:A C1#3:A
        """,
        {
            'block_from': ([(0, {"resname": 'ALA', 'resid': 1, 'atomname': 'A'}),
                            (2, {"resname": 'GLY', 'resid': 2, 'atomname': 'A'})],
                           []),
            'block_to': ([(0, {"resname": 'A', 'resid': 5, 'atomname': 'A'}),
                          (1, {"resname": 'B13', 'resid': 1, 'atomname': 'A'}),
                          (2, {"resname": 'B13', 'resid': 1, 'atomname': 'B'}),
                          (3, {"resname": 'C1', 'resid': 3, 'atomname': 'A'})],
                         []),
            'mapping': {0: {0: 1, 1: 1}, 2: {3: 1}}
        }
    ),
    (  # Block shorthand, fetch single block
        """
        [ block ]
        [ from ]
        test_A
        [ from blocks ]
        block_A
        [ to blocks ]
        !A
        [ to nodes ]
        A
        [ mapping ]
        A1 A
        A2 A
        B1 A
        B2 A
        """,
        {
            'block_from': ([(0, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'A1', 'charge_group': 1}),
                            (1, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'A2', 'charge_group': 1}),
                            (2, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'B1', 'charge_group': 1}),
                            (3, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'B2', 'charge_group': 1})],
                           [(0, 1, {}), (2, 3, {})]),
            'block_to': ([(0, {"resname": 'A', 'resid': 1, 'atomname': 'A'})],
                         []),
            'mapping': {0: {0: 1}, 1: {0: 1}, 2: {0: 1}, 3: {0: 1}}
        }
    ),
    (  # Block shorthand, fetch single modification
        """
        [ modification ]
        [ from ]
        test_A
        [ from blocks ]
        block_A
        [ from nodes ]
        C
        [ to blocks ]
        !A
        [ to nodes ]
        A
        [ mapping ]
        A1 A
        A2 A
        B1 A
        B2 A
        C  A
        """,
        {
            'block_from': ([(0, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'A1', 'charge_group': 1}),
                            (1, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'A2', 'charge_group': 1}),
                            (2, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'B1', 'charge_group': 1}),
                            (3, {"resname": 'block_A', 'resid': 1,
                                 'atomname': 'B2', 'charge_group': 1}),
                            (4, {'resid': 1, 'atomname': 'C'})],
                           [(0, 1, {}), (2, 3, {})]),
            'block_to': ([(0, {'resid': 1, 'atomname': 'A'})],
                         []),
            'mapping': {0: {0: 1}, 1: {0: 1}, 2: {0: 1}, 3: {0: 1}, 4: {0: 1}}
        }
    ),
    (  # Block longhand, fetch multiple
        """
        [ block ]
        [ to ]
        test_A
        [ to blocks ]
        block_A#1 {"resname": "block_A", "resid": 1}
        blockA {"resname": "block_A", "resid": 2}
        [ from blocks ]
        !A
        [ from nodes ]
        A
        [ mapping ]
        A block_A#1:A1
        A A2
        A B1
        A B2
        A blockA:A1
        A A2
        A B1
        A B2
        """,
        {
            'block_to': ([(0, {"resname": 'block_A', 'resid': 1,
                               'atomname': 'A1', 'charge_group': 1}),
                          (1, {"resname": 'block_A', 'resid': 1,
                               'atomname': 'A2', 'charge_group': 1}),
                          (2, {"resname": 'block_A', 'resid': 1,
                               'atomname': 'B1', 'charge_group': 1}),
                          (3, {"resname": 'block_A', 'resid': 1,
                               'atomname': 'B2', 'charge_group': 1}),
                          (4, {"resname": 'block_A', 'resid': 2,
                               'atomname': 'A1', 'charge_group': 2}),
                          (5, {"resname": 'block_A', 'resid': 2,
                               'atomname': 'A2', 'charge_group': 2}),
                          (6, {"resname": 'block_A', 'resid': 2,
                               'atomname': 'B1', 'charge_group': 2}),
                          (7, {"resname": 'block_A', 'resid': 2,
                               'atomname': 'B2', 'charge_group': 2})],
                         [(0, 1, {}), (2, 3, {}), (4, 5, {}), (6, 7, {})]),
            'block_from': ([(0, {"resname": 'A', 'resid': 1, 'atomname': 'A'})],
                           []),
            'mapping': {0: {0: 1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1}}
        }
    ),
    (  # Block shorthand, fetch single from block with replace attr
        """
        [ block ]
        [ from ]
        test_A
        [ from blocks ]
        block_B
        [ to blocks ]
        !A
        [ to nodes ]
        A
        [ mapping ]
        B A
        """,
        {
            'block_from': ([(0, {"resname": 'block_B', 'resid': 1,
                                 'atomname': 'B', 'charge_group': 1})],
                           []),
            'block_to': ([(0, {"resname": 'A', 'resid': 1, 'atomname': 'A'})],
                         []),
            'mapping': {0: {0: 1}}
        }
    ),
    (  # Block shorthand, fetch single to block with replace attr
        """
        [ block ]
        [ to ]
        test_A
        [ to blocks ]
        block_B
        [ from blocks ]
        !A
        [ from nodes ]
        A
        [ mapping ]
        A A
        """,
        {
            'block_to': ([(0, {"resname": 'block_B', 'resid': 1,
                               'atomname': 'A', 'charge_group': 1,
                               'replace': {'atomname': 'B'}})],
                         []),
            'block_from': ([(0, {"resname": 'A', 'resid': 1, 'atomname': 'A'})],
                           []),
            'mapping': {0: {0: 1}}
        }
    ),
))
def test_single_mapping_attrs(director, lines, expected):
    """
    Tests that when a file defines a single mapping the produced mapping has
    the correct attributes.
    """
    lines = textwrap.dedent(lines).splitlines()
    mappings = list(director.parse(lines))
    assert len(mappings) == 1
    mapping = mappings[0]
    for attr_name, val in expected.items():
        if 'block' in attr_name:
            nodes, edges = val
            equal_graph_data(list(getattr(mapping, attr_name).nodes(data=True)),
                             list(getattr(mapping, attr_name).edges(data=True)),
                             nodes, edges)
        else:
            assert getattr(mapping, attr_name) == val, attr_name


@pytest.mark.parametrize('map_type', (
    '[block]',
    '[modification]'
))
@pytest.mark.parametrize('number', (
    1,
    2,
    3
))
def test_multiple_mappings(director, map_type, number):
    """
    Tests that when a single file defines multiple mappings it produces the
    correct number of mappings.
    """
    lines = (map_type + '\n') * number
    lines = lines.splitlines()
    mappings = list(director.parse(lines))
    assert len(mappings) == number



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains tests for the DeferredFileWriter
"""
from pathlib import Path
import os
import pytest
from vermouth.file_writer import DeferredFileWriter


def test_is_singleton():
    """Ensure the DeferredFileWriter is a singleton"""
    writer1 = DeferredFileWriter()
    writer2 = DeferredFileWriter()

    assert writer1 is writer2


@pytest.mark.parametrize('name, existing_files, expected', [
    ('a', [], []),
    ('a', ['a'], ['#a.1#']),
    ('a.txt', ['a.txt'], ['#a.txt.1#']),
    ('a.txt', ['a'], ['a']),
    ('a.txt', ['a.txt', '#a.txt.1#'], ['#a.txt.2#', '#a.txt.1#']),
    ('a.txt', ['a.txt', '#a.txt.2#'], ['#a.txt.1#', '#a.txt.2#']),
])
def test_backup(tmpdir, monkeypatch, name, existing_files, expected):
    """
    Ensure the DeferredFileWriter backs up existing files correctly, and at the
    correct moment
    """
    monkeypatch.chdir(tmpdir)
    for idx, file in enumerate(existing_files):
        with open(file, 'w') as handle:
            handle.write(str(idx))

    writer = DeferredFileWriter()
    with writer.open(name, 'w') as handle:
        handle.write("new {}".format(name))
    writer.write()

    assert Path(name).is_file()
    with open(name) as file:
        assert file.read() == "new {}".format(name)

    for idx, name in enumerate(expected):
        assert Path(name).is_file()
        with open(name) as file:
            assert file.read() == str(idx)


def test_deferred_writing(tmpdir, monkeypatch):
    """
    Ensure the DeferredFileWriter writes changes to files at the correct moment
    """
    monkeypatch.chdir(tmpdir)

    file_name = Path('my_file.txt')
    writer = DeferredFileWriter()
    assert not file_name.exists()
    with writer.open(file_name, 'w') as file:
        file.write('hello')
    assert not file_name.exists()
    os.chdir('..')
    writer.write()
    os.chdir(str(tmpdir))
    assert file_name.exists()
    assert file_name.read_text() == 'hello'


def test_binary_writing(tmpdir, monkeypatch):
    """Ensure the DeferredFileWriter can write and append to binary files"""
    monkeypatch.chdir(tmpdir)
    file_name = Path('my_file.txt')
    writer = DeferredFileWriter()
    assert not file_name.exists()

    with writer.open(file_name, 'wb') as file:
        file.write(b'Hello')
    assert not file_name.exists()
    os.chdir('..')
    writer.write()
    os.chdir(str(tmpdir))
    assert file_name.exists()
    assert file_name.read_text() == 'Hello'

    with writer.open(file_name, 'ab') as file:
        file.write(b' world!')
    assert file_name.exists()
    assert file_name.read_text() == 'Hello'
    writer.write()
    assert file_name.read_text() == 'Hello world!'


def test_rw_plus(tmpdir, monkeypatch):
    """Ensure the DeferredFileWriter can deal with mode r+"""
    monkeypatch.chdir(tmpdir)
    path = Path('file.txt')
    path.write_text('123')
    writer = DeferredFileWriter()

    with writer.open(path, 'r+') as file:
        assert file.read() == '123'
        file.write('456')
        file.seek(0)
        assert file.read() == '123456'
    assert path.read_text() == '123'
    writer.write()

    assert path.read_text() == '123456'

@pytest.mark.parametrize('mode, exception', (
    ['o', KeyError],
))
def test_mode_errors(mode, exception):
    """
    Ensure the DeferredFileWriter raises an appropriate error for unknown file
    modes
    """
    writer = DeferredFileWriter()
    with pytest.raises(exception):
        writer.open('somefile.txt', mode)


def test_append(tmpdir, monkeypatch):
    """Ensure the DeferredFileWriter can append"""
    monkeypatch.chdir(tmpdir)
    path = Path('file.txt')
    path.write_text('123')

    writer = DeferredFileWriter()
    with writer.open(path, 'a') as file:
        file.write('abc')

    assert path.read_text() == '123'

    writer.write()

    assert path.read_text() == '123abc'


def test_closing(tmpdir, monkeypatch):
    """
    Ensure the DeferredFileWriter's close method doesn't prompt writing and
    removes any temporary files.
    """
    monkeypatch.chdir(tmpdir)
    tmpdir = Path(str(tmpdir))
    writer = DeferredFileWriter()
    monkeypatch.setattr(writer, '_tmpdir', str(tmpdir))

    assert not [p.name for p in tmpdir.iterdir()]

    with writer.open('file.txt', 'w') as file:
        file.write('abc')

    writer.write()

    assert [p.name for p in tmpdir.iterdir()] == ['file.txt']

    with writer.open('file2.txt', 'w') as file:
        file.write('abc')

    writer.close()
    assert [p.name for p in tmpdir.iterdir()] == ['file.txt']


def test_reopen(tmpdir, monkeypatch):
    monkeypatch.chdir(tmpdir)
    path = Path('file.txt')
    writer = DeferredFileWriter()

    with writer.open(path, 'w') as file:
        file.write('Hello!')
    # Close file
    with writer.open(path, 'r') as file:
        assert file.read() == 'Hello!'



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the `test_utils.py` module.
"""

import itertools
import string
import pytest
from hypothesis import strategies, given, example
import numpy as np
from numpy.testing import assert_allclose
from vermouth import utils

# pylint: disable=no-value-for-parameter

DIFFERENCE_USE_CASE = (
    '', [], 'a', 2, 3, True, False, None, float('nan'), float('inf'),
    [True], [False], [True, False], [False, False], [None], [None, None],
    [1.1, 1.1], [0.0, 1.1],
    [float('nan'), float('inf')],
    [float('inf'), float('nan')],
    ['a', 'b'], 'ab',
    np.array([1.2, 2.3, 4.5]),
    np.array([[1.2, 2.3, 4.5], [5.6, 7.8, 8.9]]),
    np.array([np.nan, np.inf, 3]),
    [[6, 8], [9, 10, 11]],
    {'a': 3, 2: 'tata'},
    {'a': 3, 2: np.array([2, 3, 4])},
    {'a': 4, 2: 'not tata'},
    -9223372036854775809,  # beyond an int64, can cause numpy to break
)



@pytest.mark.parametrize('iter_type', (list, tuple, iter, np.array))
@pytest.mark.parametrize('values, ref', (
    # All positive integers, one maximum
    ([1, 2, 3, 4], [4]),
    ([4, 2, 1, 3], [4]),
    # All positive integers, multiple maxima
    ([4, 1, 4, 2], [4] * 2),
    ([4, 4, 4, 4], [4] * 4),
    # All positive or negative integers, multiple maxima
    ([-1, 2, -4, 0, 2], [2] * 2),
    ([-1, -2, -3, -1], [-1] * 2),
    # All floats
    ([0.1, 0.4, 3.4, 2.1, 3.4], [3.4] * 2),
    # Strings
    (['z', 'aaa', 'bbb', 'z', 'z'], ['z'] * 3),
))
def test_maxes_simple_no_key(values, ref, iter_type):
    """
    Test that :func:`utils.maxes` works in the simplest case.
    """
    assert utils.maxes(iter_type(values)) == ref


@pytest.mark.parametrize('iter_type', (list, tuple, iter, np.array))
@pytest.mark.parametrize('values, ref', (
    # All positive integers, one maximum
    ([1, 2, 3, 4], [1]),
    ([4, 2, 1, 3], [1]),
    # All positive integers, multiple maxima
    ([4, 1, 1, 2], [1] * 2),
    ([4, 4, 4, 4], [4] * 4),
    # All positive or negative integers, multiple maxima
    ([-1, 2, -4, 0, -4], [-4] * 2),
    ([-3, -2, -3, -1], [-3] * 2),
    # All floats
    ([0.1, 0.4, 3.4, 0.1, 3.4], [0.1] * 2),
))
def test_maxes_key(values, ref, iter_type):
    """
    Test that :func:`utils.maxes` works when provided with a key.
    """
    assert utils.maxes(iter_type(values), key=lambda x: -x) == ref


@pytest.mark.parametrize('alpha', string.ascii_letters)
@pytest.mark.parametrize('template', (
    '{}bc', ' {}bc', '123{}WS', '@#$%{}*&^', '\U0001F62E{}',
))
def test_first_alpha(template, alpha):
    """
    Test that :func:`utils.first_alpha` works as expected.
    """
    value = template.format(alpha)
    assert utils.first_alpha(value) == alpha


@pytest.mark.parametrize('value', ('', '0123', '\U0001F62E'))
def test_first_alpha_value_error(value):
    """
    Make sure :func:`utils.first_alpha` raise a ValueError when expected to.
    """
    with pytest.raises(ValueError):
        utils.first_alpha(value)


@pytest.mark.parametrize('iter_type', (list, tuple, iter, np.array))
@pytest.mark.parametrize('values, ref', (
    # All equal
    ([0] * 5, True), ([True] * 4, True),
    ([None] * 5, True), ([1e-3] * 4, True),
    # Not all equal
    ([0, 1] * 4, False), (['a', 'b', 'c'], False),
    # Empty
    ([], True),
    # One element
    ([1], True), ([None], True),
    # Should it be this way?
    ([0, False] * 3, True),
))
def test_are_all_equal(values, ref, iter_type):
    """
    Make sure that :func:`utils.are_all_equal` works as expected.
    """
    assert utils.are_all_equal(iter_type(values)) is ref


@pytest.mark.parametrize('values, ref', (
    ([list([0, 1, 2]) for _ in range(3)], True),
    ([[0, 1, 2], [3, 4, 5], [6, 7, 8]], False),
))
def test_are_all_equal_multidim_list(values, ref):
    """
    Test :func:`utils.are_all_equal` works on nested lists.
    """
    assert utils.are_all_equal(values) is ref


def test_are_all_equal_not_implemented():
    """
    Test that :func:`utils.are_all_equal` raises an exception on
    multidimensional arrays.
    """
    value = np.zeros((3, 4))
    with pytest.raises(NotImplementedError):
        utils.are_all_equal(value)


@strategies.composite
def vector_with_random_distance(draw):
    """
    Generate a vector with a random length and orientation.

    The vector is returned as the two points at its extremity.

    Returns
    -------
    point1: numpy.ndarray
    point2: numpy.ndarray
    distance: float
    """
    # Generate random polar coordinates and convert them to euclidean
    # coordinates.
    length = strategies.floats(min_value=0, max_value=100,
                               allow_nan=False, allow_infinity=False)
    angle = strategies.floats(min_value=0, max_value=2 * np.pi,
                              allow_nan=False, allow_infinity=False)
    distance = draw(length)
    theta = draw(angle)
    phi = draw(angle)
    shift = np.array([draw(length), draw(length), draw(length)])
    x = distance * np.sin(theta) * np.cos(phi)  # pylint: disable=invalid-name
    y = distance * np.sin(theta) * np.sin(phi)  # pylint: disable=invalid-name
    z = distance * np.cos(theta)  # pylint: disable=invalid-name
    return shift, np.array([x, y, z]) + shift, distance


@given(vector_with_random_distance())
@example((np.zeros((3,)), np.zeros((3,)), 0))
def test_distance(vec_and_dist):
    """
    Test the results of :func:`utils.distance`.
    """
    point1, point2, distance = vec_and_dist
    assert_allclose(utils.distance(point1, point2), distance)


@pytest.mark.parametrize(
    'left, right',
    itertools.combinations(DIFFERENCE_USE_CASE, 2)
)
def test_are_different(left, right):
    """
    Test that :func:`are_different` identify different values.
    """
    assert utils.are_different(left, right)


@pytest.mark.parametrize('left', DIFFERENCE_USE_CASE)
def test_not_are_different(left):
    """
    Test that :func:`are_different` returns False for equal values.
    """
    assert not utils.are_different(left, left)


@pytest.mark.parametrize('left, right, expected', (
    (  # dict with different order
        {-1: None, '0+': None},
        {'0+': None, -1: None},
        False,
    ),
))
def test_are_different_custom(left, right, expected):
    """
    Test :func:`are_different` on handcrafted cases.
    """
    assert utils.are_different(left, right) == expected



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that can add edges to a graph based on geometric criteria.
"""

from collections import defaultdict
import networkx as nx
import numpy as np

from .. import KDTree
from ..molecule import Molecule
from .processor import Processor
from ..utils import format_atom_string
from ..graph_utils import collect_residues, partition_graph

from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


# Van der Waals radii from A. Bondi, J. Phys. Chem., 68, 441-452, 1964.
# https://doi.org/10.1021/j100785a001
# For hydrogen, we use R.S. Rowland & R. Taylor, J.Phys.Chem., 100, 7384-7391, 1996.
# https://doi.org/10.1021/jp953141
VDW_RADII = {  # in nm
    'H': 0.120,
    'He': 0.140,
    'C': 0.170,
    'N': 0.155,
    'O': 0.152,
    'F': 0.147,
    'Ne': 0.154,
    'Si': 0.210,
    'P': 0.180,
    'S': 0.180,
    'Cl': 0.175,
    'Ar': 0.188,
    'As': 0.185,
    'Se': 1.90,
    'Br': 0.185,
    'Kr': 0.202,
    'Te': 0.206,
    'I': 0.198,
    'Xe': 0.216,
}
#VALENCES = {'H': 1, 'C': 4, 'N': 3, 'O': 2, 'S': 6}


def _bonds_from_distance(graph, nodes=None, non_edges=None, fudge=1.2):
    """Add edges to `graph` between `nodes` based on distance.

    Adds edges to `graph` between nodes in `nodes`, but will never add an edge
    that is in `non_edges`, nor between H atoms. It will also not create edges
    where H atoms bridge separate residues. Residues are defined by the
    '_res_serial' attribute of nodes.
    Edges are added based on a simple distance criterion. The criterion can be
    adjusted using `fudge`. Nodes need to have an element attribute that is in
    VDW_RADII in order to be eligible.

    Parameters
    ----------
    graph: networkx.Graph
        Nodes in the graph must have the attributes 'element', 'position', and
        '_res_serial'.
    nodes: collections.abc.Collection[collections.abc.Hashable]
        The nodes that should be considered for making edges. Must be in
        `graph`.
    non_edges: collections.abc.Container[frozenset[collections.abc.Hashable, collections.abc.Hashable]]
        A container of pairs of node keys between which no edge should be added,
        even when they are close enough.
    fudge: float
    """
    if not nodes:
        nodes = graph.nodes
    if not non_edges:
        non_edges = set()

    # We filter out the nodes for which we do not know the radius. Indeed, we
    # consider these nodes cannot make bonds. The filtering is done before we
    # enter the KDTree; we only provide to the KDTree the position of the nodes
    # that could make a bond. `idx_to_nodenum` make the link between the
    # indices in the `positions` array, and the node keys in the `system`
    # graph.
    idx_to_nodenum = dict(enumerate(
        subn for subn in graph
        if subn in nodes and graph.nodes[subn].get('element') in VDW_RADII
    ))
    # Guard against the case where there are no atoms with known elements, which
    # max does *not* like.
    if idx_to_nodenum:
        max_dist = max(VDW_RADII[graph.nodes[idx]['element']] for idx in idx_to_nodenum.values())
    else:
        max_dist = 0
    max_dist *= fudge

    positions = np.array([
        graph.nodes[node]['position']
        for node in idx_to_nodenum
    ], dtype=float)

    # Pylint ignore because positions is a numpy array.
    if len(positions):  # pylint: disable=len-as-condition
        positions = np.atleast_2d(positions)
        tree = KDTree(positions)
        pairs = tree.sparse_distance_matrix(tree, max_dist * fudge)
    else:
        pairs = {}

    nodes = graph.nodes
    for (idx1, idx2), dist in pairs.items():
        if idx1 >= idx2:
            continue
        node_idx1 = idx_to_nodenum[idx1]
        node_idx2 = idx_to_nodenum[idx2]

        if frozenset((node_idx1, node_idx2)) in non_edges:
            continue
        atom1 = nodes[node_idx1]
        atom2 = nodes[node_idx2]
        element1 = atom1['element']
        element2 = atom2['element']
        resserial1 = atom1['_res_serial']
        resserial2 = atom2['_res_serial']

        # Forbid H-H bonds, and in addition, prevent hydrogens from making bonds
        # to different residues.
        if element1 == 'H' and element2 == 'H' or \
                (resserial1 != resserial2 and (element1 == 'H' or element2 == 'H')):
            continue

        bond_distance = 0.5 * (VDW_RADII[element1] + VDW_RADII[element2])
        if dist <= bond_distance * fudge and not graph.has_edge(node_idx1, node_idx2):
            LOGGER.debug("Guessed bond between {} and {} based on distance.",
                         format_atom_string(atom1),
                         format_atom_string(atom2))

            graph.add_edge(node_idx1, node_idx2, distance=dist)


def _bonds_from_names(graph, resname, nodes, force_field):
    """Add edges between `nodes` in `graph` based on atom names.

    Adds edges to `graph`, assuming the nodes in `nodes` constitute a residue
    with residue name `resname`, which can be found among the `force_field`
    blocks. Edges will be added as they are in the reference Block. In addition,
    all non-edges in the Block will be generated and returned.

    Parameters
    ----------
    graph: networkx.Graph
    resname: str
    nodes: collections.abc.Iterable[collections.abc.Hashable]
        Should be node keys in `graph`
    force_field: vermouth.forcefield.ForceField
        Force field in which to look for the block with name `resname`

    Raises
    ------
    KeyError
        If `resname` is not one of the blocks known to `force_field`; or when
        a residue contains duplicate atom names.

    Returns
    -------
    Set[Frozenset[collections.abc.Hashable, collections.abc.Hashable]]
        All non-edges found in the block, with node keys from `graph`.
    """
    block = force_field.blocks.get(resname)
    if not block:
        raise KeyError("Residue {} is not known to force field {}"
                       "".format(resname, force_field.name))

    mol_name_to_idx = defaultdict(set)
    for graph_idx in nodes:
        if 'atomname' in graph.nodes[graph_idx]:
            mol_name_to_idx[graph.nodes[graph_idx]['atomname']].add(graph_idx)
    mol_name_to_idx = dict(mol_name_to_idx)
    for name, graph_idxs in mol_name_to_idx.items():
        if len(graph_idxs) > 1:
            raise KeyError("Residue has multiple atoms with atom name {}"
                           "".format(name))
        mol_name_to_idx[name] = mol_name_to_idx[name].pop()

    for block_idx, block_jdx in block.edges:
        block_idx_name = block.nodes[block_idx]['atomname']
        block_jdx_name = block.nodes[block_jdx]['atomname']
        if block_idx_name in mol_name_to_idx and block_jdx_name in mol_name_to_idx:
            graph_idx = mol_name_to_idx[block_idx_name]
            graph_jdx = mol_name_to_idx[block_jdx_name]
            pos1 = np.array(graph.nodes[graph_idx].get('position', np.full(3, np.nan)))
            pos2 = np.array(graph.nodes[graph_jdx].get('position', np.full(3, np.nan)))
            dist = np.sqrt(np.sum((pos1 - pos2)**2))
            graph.add_edge(graph_idx, graph_jdx, distance=dist)

    non_edges = set()
    for block_idx, block_jdx in nx.non_edges(block):
        block_idx_name = block.nodes[block_idx]['atomname']
        block_jdx_name = block.nodes[block_jdx]['atomname']
        if block_idx_name in mol_name_to_idx and block_jdx_name in mol_name_to_idx:
            non_edges.add(frozenset((mol_name_to_idx[block_idx_name],
                                     mol_name_to_idx[block_jdx_name])))
    return non_edges


def make_bonds(system, allow_name=True, allow_dist=True, fudge=1.2):
    """Creates bonds within molecules in the system.

    First, edges will be created based on residue and atom names. Second, edges
    will be created based on a distance criterion. Nodes in system must have
    `position` and `element` attributes. The possible distance between nodes is
    determined by values in `VDW_RADII`. Edges within residues will only be
    guessed between atoms that are not known in the reference Block.
    The system will be split into connected components, keeping residues
    (identified by chain, residue name and residue id) within the same molecule.
    This does mean that the final molecules can be disconnected.

    Notes
    -----
    Edges for residues for which no block can be found will be added based on
        the distance criterion. A warning will be issued if this is the case.

    Elements that are not in `VDW_RADII` do not make bonds based on distances.

    Parameters
    ----------
    system: :class:`~vermouth.system.System`
        The system in which to add edges.
    fudge: :class:`~numbers.Number`
        Scale the allowed distance by this factor.

    Returns
    -------
    List[:class:`~vermouth.molecule.Molecule`]
        Molecules in system, in which edges have been added based on atom names
        and possibly distance. The molecules have been split into connected
        components keeping residues intact. Molecules can be disconnected within
        residues.
    """
    force_field = system.force_field

    # Separate molecules should remain separate molecules, even if they have
    # poorly chosen chain/resname/resid combinations. So add a mol_idx attribute
    for mol_idx, molecule in enumerate(system.molecules):
        nx.set_node_attributes(molecule, mol_idx, 'mol_idx')

    system = nx.disjoint_union_all(system.molecules)
    non_edges = set()

    residue_groups = collect_residues(system, ('mol_idx chain resid resname insertion_code'.split()))
    for res_serial, (keys, idxs) in enumerate(residue_groups.items()):
        mol_idx, chain, resid, resname, insertion_code = keys
        for idx in idxs:
            system.nodes[idx]['_res_serial'] = res_serial
        if not allow_name:
            continue
        try:
            # Try adding bonds within the residue based on atom names
            non_edges.update(_bonds_from_names(system, resname, idxs, force_field))
        except KeyError as error:
            # ... if that doesn't work, fall back to distance
            message = "Can't add bonds based on atom names for residue {}-{}{} because {}."
            if allow_dist:
                _bonds_from_distance(system, idxs, fudge=fudge)
                message += " Falling back to distance criteria."
            LOGGER.warning(message,
                           chain, resname, resid, error, force_field.name)
    # And finally, add edges based on distance, but ignore any edges that would
    # otherwise have been added by name. So only edges between residues will be
    # added, and edges involving atoms which are not known to the blocks (PTMs,
    # termini, ...)
    if allow_dist:
        _bonds_from_distance(system, non_edges=non_edges, fudge=fudge)
    # Split the system into connected components. We do want to keep residues
    # together, so make a residue graph (1 node per residue) first, and use that
    # to find the connected components.
    molecules = []
    residue_graph = partition_graph(system, residue_groups.values())
    for res_node_idxs in nx.connected_components(residue_graph):
        node_idxs = set().union(*(residue_graph.nodes[rni]['graph'] for rni in res_node_idxs))
        mol = Molecule(system.subgraph(node_idxs))
        molecules.append(mol)

    return molecules


class MakeBonds(Processor):
    def __init__(self, allow_name=True, allow_dist=True, fudge=1.2):
        self.allow_name = allow_name
        self.allow_dist = allow_dist
        self.fudge = fudge

    def run_system(self, system):
        if not system.molecules:
            # No molecules means nothing to do.
            return
        mols = make_bonds(system,
                          allow_name=self.allow_name,
                          allow_dist=self.allow_dist,
                          fudge=self.fudge)
        system.molecules = mols
        # Restore the force field in each molecule. Setting the force field
        # at the system level propagates it to all the molecules.
        system.force_field = system.force_field
        LOGGER.info('{} molecules after guessing bonds', len(system.molecules))



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the :mod:`vermouth.processors.average_beads` module.
"""

import pytest
import networkx as nx
import numpy as np

from vermouth.processors import average_beads

# pylint: disable=redefined-outer-name

@pytest.fixture
def mol_with_subgraph():
    """
    Create a molecule with a subgraph under the "graph" attribute of each node.
    """
    mol = nx.OrderedGraph()

    subgraph = nx.Graph()
    subgraph.add_nodes_from((
        (0, {'mass': 1.1, 'not mass': 2.1, 'position': np.array([1, 2, 3], dtype=float),}),
        (1, {'mass': 1.2, 'not mass': 2.2, 'position': np.array([2, 3, 4], dtype=float),}),
        (2, {'mass': 1.3, 'not mass': 2.3, 'position': np.array([3, 4, 5], dtype=float),}),
    ))
    mol.add_node(0, **{
        "graph": subgraph,
        "mapping_weights": {0: 1, 1: 2, 2: 3},
        "target mass": np.array([2.378378378, 3.378378378, 4.378378378]),
        "target not mass": np.array([2.358208955, 3.358208955, 4.358208955]),
        "target None": np.array([2.333333333, 3.333333333, 4.333333333]),
        "target False": np.array([2.333333333, 3.333333333, 4.333333333]),
    })


    subgraph = nx.Graph()
    subgraph.add_nodes_from((
        (0, {'mass': 1.2, 'not mass': 2.2, 'position': np.array([2, 3, 4], dtype=float),}),
        (1, {'mass': 1.3, 'not mass': 2.3, 'position': np.array([3, 4, 5], dtype=float),}),
    ))
    mol.add_node(1, **{
        "graph": subgraph,
        "mapping_weights": {0: 2, 1: 3},
        "target mass": np.array([2.619047619, 3.619047619, 4.619047619]),
        "target not mass": np.array([2.610619469, 3.610619469, 4.610619469]),
        "target None": np.array([2.6, 3.6, 4.6]),
        "target False": np.array([2.6, 3.6, 4.6]),
    })

    subgraph = nx.Graph()
    mol.add_node(2, **{
        "graph": subgraph,
        "mapping_weights": {},
        "target mass": np.array([np.nan, np.nan, np.nan]),
        "target not mass": np.array([np.nan, np.nan, np.nan]),
        "target None": np.array([np.nan, np.nan, np.nan]),
        "target False": np.array([np.nan, np.nan, np.nan]),
    })

    subgraph = nx.Graph()
    subgraph.add_nodes_from((
        (0, {'mass': 1.2, 'not mass': 2.2, 'position': np.array([2, 3, 4], dtype=float),}),
        (1, {'mass': 1.3, 'not mass': 2.3, 'position': np.array([3, 4, 5], dtype=float),}),
    ))
    mol.add_node(3, **{
        "graph": subgraph,
        "mapping_weights": {0: 0, 1: 0},
        "target mass": np.array([np.nan, np.nan, np.nan]),
        "target not mass": np.array([np.nan, np.nan, np.nan]),
        "target None": np.array([np.nan, np.nan, np.nan]),
        "target False": np.array([np.nan, np.nan, np.nan]),
    })

    return mol

@pytest.fixture(params=(None, 'mass', 'not mass'))
def mol_with_variable(request, mol_with_subgraph):
    """
    Build a mock molecule with a mock force field declaring 'center_weight'.
    """
    weight = request.param

    class MockForceField:
        pass

    ff = MockForceField()
    ff.variables = {'center_weight': weight}

    mol_with_subgraph.force_field = ff
    return mol_with_subgraph


@pytest.mark.parametrize('weight', (None, 'mass', 'not mass'))
def test_do_average_bead(mol_with_subgraph, weight):
    """
    Test normal operation of :func:`average_beads.do_average_bead`.
    """
    result_mol = average_beads.do_average_bead(
        mol_with_subgraph, ignore_missing_graphs=False, weight=weight,
    )
    target_key = 'target {}'.format(weight)
    target_positions = np.stack([node[target_key] for node in mol_with_subgraph.nodes.values()])
    positions = np.stack([node['position'] for node in mol_with_subgraph.nodes.values()])
    assert np.allclose(positions, target_positions, equal_nan=True)


@pytest.mark.parametrize('weight', ('mass', 'not mass'))
def test_shoot_weight(mol_with_subgraph, weight):
    """
    Test that :func:`average_beads.do_average_bead` fails if a weight is missing.
    """
    del mol_with_subgraph.nodes[0]['graph'].nodes[1][weight]
    with pytest.raises(KeyError):
        average_beads.do_average_bead(
            mol_with_subgraph, ignore_missing_graphs=False, weight=weight,
        )


def test_shoot_graph(mol_with_subgraph):
    """
    Test that :func:`average_beads.do_average_bead` fails if a subgraph is missing.
    """
    del mol_with_subgraph.nodes[1]['graph']
    with pytest.raises(ValueError):
        average_beads.do_average_bead(mol_with_subgraph)


def test_processor_variable(mol_with_variable):
    processor = average_beads.DoAverageBead()
    mol = processor.run_molecule(mol_with_variable)
    weight = mol_with_variable.force_field.variables['center_weight']
    target_key = 'target {}'.format(weight)
    target_positions = np.stack([node[target_key] for node in mol_with_variable.nodes.values()])
    positions = np.stack([node['position'] for node in mol_with_variable.nodes.values()])
    assert np.allclose(positions, target_positions, equal_nan=True)


@pytest.mark.parametrize('weight', (False, 'mass', 'not mass'))
def test_processor_weight(mol_with_variable, weight):
    processor = average_beads.DoAverageBead(weight=weight)
    mol = processor.run_molecule(mol_with_variable)
    target_key = 'target {}'.format(weight)
    target_positions = np.stack([node[target_key] for node in mol_with_variable.nodes.values()])
    positions = np.stack([node['position'] for node in mol_with_variable.nodes.values()])
    assert np.allclose(positions, target_positions, equal_nan=True)



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
VerMoUTH: The Very Modular Universal Transformation Helper

Provides functionality for creating MD topologies from coordinate files. Powers
the CLI tool martinize2.
"""
import logging

import pbr.version

from .log_helpers import StyleAdapter, get_logger

__version__ = pbr.version.VersionInfo('vermouth').release_string()

logging.getLogger(__name__).addHandler(logging.NullHandler())

LOGGER = StyleAdapter(get_logger(__name__))


# Find the data directory once.
try:
    import pkg_resources
except ImportError:
    import os
    DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
    del os
else:
    DATA_PATH = pkg_resources.resource_filename('vermouth', 'data')
    del pkg_resources

del pbr

try:
    from scipy.spatial import cKDTree as KDTree
except ImportError:
    LOGGER.info('Using redistributed KDTree. Some functionality might be slow.'
                ' Install scipy for better performance.', type='performance')
    from .redistributed.kdtree import KDTree

del LOGGER

from .molecule import Molecule  # pylint: disable=wrong-import-position
from .processors import *  # pylint: disable=wrong-import-position
from .system import System  # pylint: disable=wrong-import-position



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that adds a rubber band elastic network.
"""
import itertools

import numpy as np
import networkx as nx

from .processor import Processor
from .. import selectors
from ..graph_utils import make_residue_graph

# the bond type of the RB
DEFAULT_BOND_TYPE = 6
# the minimum distance between the resids
# of two beads to have an RB
DEFAULT_RMD = 2

def self_distance_matrix(coordinates):
    """
    Compute a distance matrix between points in a selection.

    Notes
    -----
    This function does **not** account for periodic boundary conditions.

    Parameters
    ----------
    coordinates: numpy.ndarray
        Coordinates of the points in the selection. Each row must correspond
        to a point and each column to a dimension.

    Returns
    -------
    numpy.ndarray
    """
    return np.sqrt(
        np.sum(
            (coordinates[:, np.newaxis, :] - coordinates[np.newaxis, :, :]) ** 2,
            axis=-1)
    )


def compute_decay(distance, shift, rate, power):
    r"""
    Compute the decay function of the force constant as function to the distance.

    The decay function for the force constant is defined as:

    .. math::

        \exp^{-r(d - s)^p}

    where :math:`r` is the decay rate given by the 'rate' argument,
    :math:`p` is the decay power given by 'power', :math:`s` is a shift
    given by 'shift', and :math:`d` is the distance between the two atoms given
    in 'distance'. If the rate or the power are set to 0, then the decay
    function does not modify the force constant.

    The 'distance' argument can be a scalar or a numpy array. If it is an
    array, then the returned value is an array of decay factors with the same
    shape as the input.
    """
    return np.exp(-rate * ((distance - shift) ** power))


def compute_force_constants(distance_matrix, lower_bound, upper_bound,
                            decay_factor, decay_power, base_constant,
                            minimum_force):
    """
    Compute the force constant of an elastic network bond.

    The force constant can be modified with a decay function, and it can be
    bounded with a minimum threshold, or a distance upper and lower bonds.
    """
    constants = compute_decay(distance_matrix, lower_bound, decay_factor, decay_power)
    np.fill_diagonal(constants, 0)
    constants *= base_constant
    constants[constants < minimum_force] = 0
    constants[distance_matrix > upper_bound] = 0
    return constants


def are_connected(graph, left, right, separation):
    """
    ``True`` if the nodes are at most 'separation' nodes away.

    Parameters
    ----------
    graph: networkx.Graph
        The graph/molecule to work on.
    left:
        One node key from the graph.
    right:
        One node key from the graph.
    separation: int
        The maximum number of nodes in the shortest path between two nodes of
        interest for these two nodes to be considered connected. Must be >= 0.

    Returns
    -------
    bool
    """
    nodes_are_connected = False
    try:
        shortest_path = len(nx.shortest_path(graph, left, right))
    except nx.NetworkXNoPath:
        # There is no path between left and right so they are not
        # connected; which is the default.
        pass
    else:
        # The source and the target are counted in the shortest path
        nodes_are_connected = shortest_path <= separation + 2
    return nodes_are_connected


def build_connectivity_matrix(graph, separation, node_to_idx, selected_nodes):
    """
    Build a connectivity matrix based on the separation between nodes in a graph.

    The connectivity matrix is a symmetric boolean matrix where cells contain
    ``True`` if the corresponding atoms are connected in the graph and
    separated by less or as much nodes as the given 'separation' argument.

    In the following examples, the separation between A and B is 0, 1, and 2.
    respectively:

    ```
    A - B
    A - X - B
    A - X - X - B
    ```

    Note that building the connectivity matrix with a separation of 0 is the
    same as building the adjacency matrix.

    Parameters
    ----------
    graph: networkx.Graph
        The graph/molecule to work on.
    separation: int
        The maximum number of nodes in the shortest path between two nodes of
        interest for these two nodes to be considered connected. Must be >= 0.
    selected_nodes: collections.abc.Collection
        A list of nodes to work on.

    Returns
    -------
    numpy.ndarray
        A boolean matrix.
    """
    res_graph = make_residue_graph(graph)
    distance_pairs = nx.all_pairs_shortest_path_length(res_graph, cutoff=separation)
    # only gets "positive" entries due to the cutoff argument above
    size = graph.number_of_nodes()
    # the matrix will be reduced before returning it but nx.all_pairs_shortest_path_length
    # does not take a subset of nodes
    # TODO optimize me,  try to create a sparse matrix in case scipy is available.
    connectivity = np.zeros((size, size), dtype=bool)
    for origin_residue, matchs_distances in distance_pairs:
        for target_residue in matchs_distances:
            origin_nodes = res_graph.nodes[origin_residue]['graph'].nodes()
            target_nodes = res_graph.nodes[target_residue]['graph'].nodes()
            for origin, target in itertools.product(origin_nodes, target_nodes):
                connectivity[node_to_idx[origin], node_to_idx[target]] = True
    np.fill_diagonal(connectivity, False)
    return connectivity[:, selected_nodes][selected_nodes]


def build_pair_matrix(graph, criterion, idx_to_node, selected_nodes):
    """
    Build a boolean matrix telling if a pair of nodes fulfil a criterion.

    Parameters
    ----------
    graph: networkx.Graph
        The graph/molecule to work on.
    criterion: collections.abc.Callable
        A function that determines if a pair of nodes fulfill the criterion.
        It takes a graph and two node keys as arguments and returns a boolean.
    selected_nodes: collections.abc.Collection
        A list of nodes to work on.

    Returns
    -------
    numpy.ndarray
        A boolean matrix.
    """
    size = len(graph.nodes)
    #TODO generate spare matrix with scipy
    share_domain = np.zeros((size, size), dtype=bool)
    node_combinations = itertools.combinations(selected_nodes, 2)
    for kdx, jdx in node_combinations:
        key_kdx = idx_to_node[kdx]
        key_jdx = idx_to_node[jdx]
        share_domain[kdx, jdx] = criterion(graph, key_kdx, key_jdx)
        share_domain[jdx, kdx] = share_domain[kdx, jdx]
    return share_domain[:, selected_nodes][selected_nodes]

def apply_rubber_band(molecule, selector,
                      lower_bound, upper_bound,
                      decay_factor, decay_power,
                      base_constant, minimum_force,
                      bond_type, domain_criterion, res_min_dist):
    r"""
    Adds a rubber band elastic network to a molecule.

    The elastic network is applied as bounds between the atoms selected by the
    function declared with the 'selector' argument. The equilibrium length for
    the bonds is measured from the coordinates in the molecule, the force
    constant is computed from the base force constant and an optional decay
    function.

    The decay function for the force constant is defined as:

    .. math::

        \exp^{-r(d - s)^p}

    where :math:`r` is the decay rate given by the 'decay_factor' argument,
    :math:`p` is the decay power given by 'decay_power', :math:`s` is a shift
    given by 'lower_bound', and :math:`d` is the distance between the two atoms
    in the molecule. If the rate or the power are set to 0, then the decay
    function does not modify the force constant.

    The 'selector' argument takes a callback that accepts a atom dictionary and
    returns ``True`` if the atom match the conditions to be kept.

    Only nodes that are in the same domain can be connected by the elastic
    network. The 'domain_criterion' argument accepts a callback that determines
    if two nodes are in the same domain. That callback accepts a graph and two
    node keys as argument and returns whether or not the nodes are in the same
    domain as a boolean.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to which apply the elastic network. The molecule is
        modified in-place.
    selector: collections.abc.Callable
        Selection function.
    lower_bound: float
        The minimum length for a bond to be added, expressed in
        nanometers.
    upper_bound: float
        The maximum length for a bond to be added, expressed in
        nanometers.
    decay_factor: float
        Parameter for the decay function.
    decay_power: float
        Parameter for the decay function.
    base_constant: float
        The base force constant for the bonds in :math:`kJ.mol^{-1}.nm^{-2}`.
        If 'decay_factor' or 'decay_power' is set to 0, then it will be the
        used force constant.
    minimum_force: float
        Minimum force constant in :math:`kJ.mol^{-1}.nm^{-2}` under which bonds
        are not kept.
    bond_type: int
        Gromacs bond function type to apply to the elastic network bonds.
    domain_criterion: collections.abc.Callable
        Function to establish if two atoms are part of the same domain. Elastic
        bonds are only added within a domain. By default, all the atoms in
        the molecule are considered part of the same domain. The function
        expects a graph (e.g. a :class:`~vermouth.molecule.Molecule`) and two atom node keys as
        argument and returns ``True`` if the two atoms are part of the same
        domain; returns ``False`` otherwise.
    res_min_dist: int
        Minimum separation between two atoms for a bond to be kept.
        Bonds are kept is the separation is greater or equal to the value
        given.
    """
    selection = []
    coordinates = []
    missing = []
    node_to_idx = {}
    idx_to_node = {}
    for node_idx, (node_key, attributes) in enumerate(molecule.nodes.items()):
        node_to_idx[node_key] = node_idx
        idx_to_node[node_idx] = node_key
        if selector(attributes):
            selection.append(node_idx)
            coordinates.append(attributes.get('position'))
            if coordinates[-1] is None:
                missing.append(node_key)
        node_idx += 1
    if missing:
        raise ValueError('All atoms from the selection must have coordinates. '
                         'The following atoms do not have some: {}.'
                         .format(' '.join(missing)))
    coordinates = np.stack(coordinates)
    distance_matrix = self_distance_matrix(coordinates)
    constants = compute_force_constants(distance_matrix, lower_bound,
                                        upper_bound, decay_factor, decay_power,
                                        base_constant, minimum_force)

    connected = build_connectivity_matrix(molecule, res_min_dist, node_to_idx,
                                          selected_nodes=selection)

    same_domain = build_pair_matrix(molecule, domain_criterion, idx_to_node,
                                    selected_nodes=selection)

    can_be_linked = (~connected) & same_domain
    # Multiply the force constant by 0 if the nodes cannot be linked.
    constants *= can_be_linked
    distance_matrix = distance_matrix.round(5)  # For compatibility with legacy
    for from_idx, to_idx in zip(*np.triu_indices_from(constants)):
        # note the indices in the matrix are not anymore the idx of
        # the full molecule but the subset of nodes in selection
        from_key = idx_to_node[selection[from_idx]]
        to_key = idx_to_node[selection[to_idx]]
        force_constant = constants[from_idx, to_idx]
        length = distance_matrix[from_idx, to_idx]
        if force_constant > minimum_force:
            molecule.add_interaction(
                type_='bonds',
                atoms=(from_key, to_key),
                parameters=[bond_type, length, force_constant],
                meta={'group': 'Rubber band'},
            )


def always_true(*args, **kwargs):  # pylint: disable=unused-argument
    """
    Returns ``True`` whatever the arguments are.
    """
    return True


def same_chain(graph, left, right):
    """
    Returns ``True`` is the nodes are part of the same chain.

    Nodes are considered part of the same chain if they both have the same value
    under the "chain" attribute, or if neither of the 2 nodes have that attribute.

    Parameters
    ----------
    graph: networkx.Graph
        A graph the nodes are part of.
    left:
        A node key in 'graph'.
    right:
        A node key in 'graph'.

    Returns
    -------
    bool
        ``True`` if the nodes are part of the same chain.
    """
    node_left = graph.nodes[left]
    node_right = graph.nodes[right]
    return node_left.get('chain') == node_right.get('chain')


class ApplyRubberBand(Processor):
    def __init__(self, lower_bound, upper_bound, decay_factor, decay_power,
                 base_constant, minimum_force,
                 res_min_dist=None,
                 bond_type=None,
                 selector=selectors.select_backbone,
                 bond_type_variable='elastic_network_bond_type',
                 res_min_dist_variable='elastic_network_res_min_dist',
                 domain_criterion=always_true):
        super().__init__()
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.decay_factor = decay_factor
        self.decay_power = decay_power
        self.base_constant = base_constant
        self.minimum_force = minimum_force
        self.bond_type = bond_type
        self.selector = selector
        self.bond_type_variable = bond_type_variable
        self.domain_criterion = domain_criterion
        self.res_min_dist = res_min_dist
        self.res_min_dist_variable = res_min_dist_variable

    def run_molecule(self, molecule):
        # Choose the bond type. From high to low, the priority order is:
        # * what is set as an argument to the processor
        # * what is written in the force field variables
        #   under the key `self.bond_type_variable`
        # * the default value set in DEFAULT_BOND_TYPE
        bond_type = self.bond_type
        if self.bond_type is None:
            bond_type = molecule.force_field.variables.get(self.bond_type_variable,
                                                           DEFAULT_BOND_TYPE)

        # Same procedure for res_min_dist the minimum distance between
        # the resids of two beads for them to have a RB
        res_min_dist = self.res_min_dist
        if self.res_min_dist is None:
            res_min_dist = molecule.force_field.variables.get(self.res_min_dist_variable,
                                                              DEFAULT_RMD)

        apply_rubber_band(molecule, self.selector,
                          lower_bound=self.lower_bound,
                          upper_bound=self.upper_bound,
                          decay_factor=self.decay_factor,
                          decay_power=self.decay_power,
                          base_constant=self.base_constant,
                          minimum_force=self.minimum_force,
                          bond_type=bond_type,
                          domain_criterion=self.domain_criterion,
                          res_min_dist=res_min_dist)
        return molecule



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that renames residues based on their current residue names
and identified modifications, such as PTMs.
"""

from .processor import Processor


def rename_modified_residues(mol):
    """
    Renames residue names based on the current residue name, and the found
    modifications. The new names are found in
    `force_field.renamed_residues`, which should be a mapping of
    ``{(rename, [modification_name, ...]): new_name}``.

    Parameters
    ----------
    mol : Molecule
        The molecule whose residue names should be changed. Is modified
        in-place.
    """
    rename_map_ff = mol.force_field.renamed_residues
    rename_map = {}
    # Sort the list of modifications, so that the order does not matter. Don't
    # make a frozenset, because it might be possible to have the same
    # modification multiple times?
    # This should probably be done as the list is parsed. Also, as we parse it,
    # make sure we actually recognize all the modification names.
    for (resname, mods), new_name in rename_map_ff.items():
        rename_map[resname, tuple(sorted(mods))] = new_name

    for node_key in mol:
        node = mol.nodes[node_key]
        modifications = node.get('modifications', [])
        resname = node.get('resname', '')
        if not modifications:
            continue
        modifications = tuple(sorted(mod.graph['name'] for mod in modifications))
        try:
            new_name = rename_map[(resname, modifications)]
        except KeyError:
            # We don't know how to rename this residue, so continue to the next
            # node.
            continue
        node['resname'] = new_name


class RenameModifiedResidues(Processor):
    def run_molecule(self, molecule):
        rename_modified_residues(molecule)
        return molecule



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .processor import Processor


def apply_posres(molecule, selector, force_constant, functype=1, ifdef='POSRES'):
    for key, node in molecule.nodes.items():
        if selector(node):
            parameters = [functype, ] + [force_constant, ] * 3
            if ifdef is not None:
                meta = {'ifdef': ifdef}
            else:
                meta = {}
            molecule.add_interaction('position_restraints',
                                     (key, ), parameters, meta)
    return molecule


class ApplyPosres(Processor):
    def __init__(self, selector, force_constant, functype=1, ifdef='POSRES'):
        super().__init__()
        self.selector = selector
        self.force_constant = force_constant
        self.functype = functype
        self.ifdef = ifdef

    def run_molecule(self, molecule):
        return apply_posres(
            molecule,
            self.selector,
            self.force_constant,
            functype=self.functype,
            ifdef=self.ifdef,
        )



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test for the NameMolType processor.
"""

# Pylint complains about hypothesis strategies not receiving a value for the
# `draw` parameter. This is because the `draw` parameter is implicitly filled
# by hypothesis. The pylint warning is disabled for the file instead of being
# disabled at every call of a strategy.
# pylint: disable=no-value-for-parameter

from collections import defaultdict
import itertools
from glob import glob
import os.path
import subprocess
import sys
import pytest
from hypothesis import given, assume
from hypothesis import strategies as st
from vermouth import System
from vermouth.processors.name_moltype import NameMolType
from .molecule_strategies import random_molecule
from .datafiles import PDB_HB
from .helper_functions import find_in_path

@st.composite
def molecules_and_moltypes(draw, max_moltypes=4, min_size=0, max_size=None,
                           molecule_kwargs=None):
    """
    Generates a list of molecules and the list of the expected moltypes.

    Parameters
    ----------
    draw:
        Internal for hypothesis.
    max_moltypes: int
        Maximum number of different moltypes.
    min_size: int
        Minimum number of molecules.
    max_size: int, optional
        Maximum number of molecules. If `None`, behaves the same way as
        :func:`hypothesis.strategies.lists`.
    molecule_kwargs: dict, optional
        Arguments to pass to :func:`random_molecule`.

    Returns
    -------
    hypothesis.searchstrategy.lazy.LazyStrategy
        A list of molecules and a list of moltype names corresponding to these
        molecules as they are expected to be assigned by
        :class:`vermouth.processors.name_moltype.NameMolType`.
    """
    if max_size is not None and max_size < min_size:
        raise ValueError('max_size ({}) must be greater than min_size ({}), or None.'
                         .format(max_size, min_size))
    if molecule_kwargs is None:
        molecule_kwargs = {}

    # We first generate the list of expected moltypes, we can then use it as
    # keys to draw the molecules with shared strategies. The moltypes must
    # match what NameMolType would generate, so they must have the form
    # molecule_0, molecule_1, molecule_x, where the index is increasing based
    # on the first occurence of the molecule type.
    # pre_moltypes will be the moltype names before we rename them based on
    # their order.
    n_moltypes = draw(st.integers(min_value=1, max_value=max_moltypes))
    unique_pre_moltypes = range(n_moltypes)
    pre_moltypes_sampled = draw(st.lists(st.sampled_from(unique_pre_moltypes),
                                         min_size=min_size, max_size=max_size))
    if not pre_moltypes_sampled:
        return [], []

    moltype_index = -1  # Will be 0 for the first moltype
    pre_to_moltype = {}
    moltypes = []
    for pre_moltype in pre_moltypes_sampled:
        if pre_moltype not in pre_to_moltype:
            moltype_index += 1
            pre_to_moltype[pre_moltype] = 'molecule_{}'.format(moltype_index)
        moltypes.append(pre_to_moltype[pre_moltype])

    sharing = defaultdict(lambda: draw(random_molecule(**molecule_kwargs)))
    molecules = [sharing[moltype] for moltype in moltypes]

    # So far we made sure that molecules with a different moltype are drawn
    # separately. Though, two different draws may result in equal molecules,
    # especially as molecules are reduced when shrinking the draws.
    representatives = [
        next(group)[0]
        for _, group in itertools.groupby(
            zip(molecules, moltypes), key=lambda x: x[1]
        )
    ]
    assume(all(not mol1.share_moltype_with(mol2)
               for mol1, mol2 in itertools.combinations(representatives, 2)))

    # Each molecule must be its own instance as molecules are modified in place.
    molecules = [molecule.copy() for molecule in molecules]

    return [molecules, moltypes]


@pytest.mark.parametrize('deduplicate', (True, False))
@given(mols_and_moltypes=molecules_and_moltypes(
    max_size=4, max_moltypes=3, molecule_kwargs={'max_nodes': 3, 'max_meta': 3},
))
def test_name_moltype(mols_and_moltypes, deduplicate):
    """
    The NameMolType processor works as expected with and withour deduplication.
    """
    molecules, moltypes = mols_and_moltypes

    system = System()
    system.molecules = molecules

    if not deduplicate:
        moltypes = ['molecule_{}'.format(i) for i in range(len(molecules))]

    processor = NameMolType(deduplicate=deduplicate)
    processor.run_system(system)

    found_moltypes = [molecule.meta['moltype'] for molecule in system.molecules]
    assert found_moltypes == moltypes


@pytest.mark.parametrize('deduplicate', (True, False))
def test_martinize2_moltypes(tmpdir, deduplicate):
    """
    Run martinize2 and make sure the ITP file produced have the expected names.
    """
    martinize2 = find_in_path()

    command = [
        sys.executable,
        martinize2,
        '-f', str(PDB_HB),
        '-o', 'topol.top',
        '-x', 'out.pdb',
        '-ignore', 'HOH', '-ignore', 'HEME',
    ]
    if deduplicate:
        expected = ['molecule_{}.itp'.format(i) for i in range(2)]
        n_outputs = 2
    else:
        command.append('-sep')
        n_outputs = 4
    expected = ['molecule_{}.itp'.format(i) for i in range(n_outputs)]

    proc = subprocess.run(command, cwd=str(tmpdir), timeout=90, check=False)
    assert proc.returncode == 0

    itp_files = sorted(os.path.basename(fname) for fname in glob(str(tmpdir / '*.itp')))
    assert itp_files == expected



#!/usr/bin/env python3
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
GRAPh PArser

Build graphs from string definition:

Grappa allows constructing a graph using a linear sequence of tokens,
the grappa string. A token can be a node name, a definition of a series
of nodes or a special character/directive. The principle behind the
grappa string is that the graph is extended from an active node. A new
node is connected to the active node. Using @ the active node can be
switched, and using () branches can be specified on an active node. A
complete description of the grappa minilanguage is given below:

  Grappa string Rules:

    name        : add node with name, with edge to active node
                  (none at start, active parent at start of branch)
    -name       : remove node with name

    @name       : select name as active node

    (           : set active node as active parent (start branching)

    ,           : switch to new branch at active parent

    )           : set active parent to active node

    =nameB      : rename active node (keep edges)

    {attr=val}  : set attribute on active node (can be
                  attributes like:  element, charge, valence, stubs
                  element is set to FIRST LETTER of name,
                  unless specified as attribute
                  attribute chiral has tuple of three nodes,
                  which define chirality according to right-hand rule

    !X          : connect active node to node X, which _must_ be present
                  already. Otherwise, using a name that is already there is an
                  error

    <NAME>      : include brick with given name

    <':NAME>    : include brick with given name and add ' as suffix to nodes

    <NAME@X>    : include brick with given name and add edge between active
                  node and node 'X' of brick

    /#=1-20/C#(H#[1-2])/ : Expand by repetion and substitution according to
                           range specified

    (/#=A-D/C#(H#[1-3]),/) : Expand to multiple branches
"""

import string

import networkx as nx

from ..log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))


DIRECTIVES = '@(),-=!'
GROUPS = ('{}', '<>')
TGROUP = '[]'


class GrappaSyntaxError(Exception):
    """Syntax of grappa string was invalid"""


def find_matching(symbols, string):
    """Find matching symbol in a series with possible nesting."""
    nesting = 0
    pos = 0
    while pos < len(string):
        if string[pos] == symbols[0]:
            nesting += 1
        elif string[pos] == symbols[1]:
            nesting -= 1
        pos += 1
        if not nesting:
            break
    else:
        raise GrappaSyntaxError("Matching '{}' not found".format(symbols[1]))
    return string[:pos]


def parse_attribute_token(attr):
    """Read attribute token and return corresponding node dictionary"""
    out = {}
    tok = tokenize(attr[1:-1], special=':,',
                   groups=('[]', '{}', '()'), tgroup=None)
    for token in tok:
        assert next(tok) == ':'
        out[token] = next(tok)
    return out


def expand_nodestring(nodestr):
    """
    Parse a string like X[1-3,6,8] to list [X1,X2,X3,X6,X8].
    X[20-25] is invalid, but X[w-z] is valid.
    """

    if '[' not in nodestr:
        return [nodestr]

    openbra = nodestr.find('[')
    closebra = nodestr.rfind(']')
    if closebra == -1:
        err = 'Matching square bracket not found in node list definition ({})'
        raise GrappaSyntaxError(err.format(nodestr))

    base = nodestr[:openbra]
    nodes = []
    what = [item.split('-') for item in nodestr[openbra+1:closebra].split(',')]
    for thing in what:
        if len(thing) == 1:
            nodes.append(thing[0].strip())
        # And here is why X[20-25] is invalid
        elif len(thing[0]) == 1 and len(thing[1]) == 1:
            for val in range(ord(thing[0]), ord(thing[1]) + 1):
                nodes.append(base + chr(val))
        else:
            err = 'Malformed range in node string ({}-{})'
            raise GrappaSyntaxError(err.format(*thing))

    return nodes


def include_graph(graphs, tag):
    """
    Prepare include graph from graphs, according to tag
    """
    atpos = tag.rfind('@')
    if atpos > 0:
        tag, atpos = tag[:atpos], tag[atpos+1:]
    else:
        atpos = ""
    lbl = tag.find(':')
    if lbl > 0:
        lbl, tag = tag[:lbl], tag[lbl+1:]
    else:
        lbl = ""
    try:
        G = graphs[tag]
    except KeyError:
        raise KeyError('Include graph {} not found in graphs.'.format(tag))
    mapping = {k: k+lbl for k in G.nodes}
    G = nx.relabel_nodes(G, mapping)
    return G, atpos + lbl


def preprocess(graphstring):
    """
    Expand graphstring 'macros'.

    EXAMPLES:

        /#=1-3/(C#(H#[1-2]))/
          --> C1(H11,H12) C2(H21,H22) C3(H31,H32)

        H1 C1(/#=2-4/(C#(H#[1-3]),)/)
          --> H1 C1(C2(H21,H22,H23),C3(H31,H32,H33),C4(H41,H42,H43))
    """
    tokkie = tokenize(graphstring, special="/", groups=(), tgroup=None)
    out = []
    for token in tokkie:
        if token == '/':
            rng = next(tokkie)

            # The range should be simple (single), so the next token must be '/'
            nexttok = next(tokkie)
            if nexttok != '/':
                err = 'Matching "/" not found during preprocessing (found {})'.format(nexttok)
                raise GrappaSyntaxError(err)

            subst, values = rng.split('=')
            values = [val.split('-') for val in values.split(',')]

            form = [next(tokkie)]

            nexttok = next(tokkie)
            try:
                while nexttok != '/':
                    form.append(nexttok)
                    nexttok = next(tokkie)
            except StopIteration:
                err = 'Matching "/" not found during preprocessing (found {})'.format(nexttok)
                raise GrappaSyntaxError(err)
            
            form = " ".join(form)

            for val in values:
                if len(val) == 1:
                    out.append(form.replace(subst, val[0]))
                elif len(val[0]) == 1 and len(val[1]) == 1:
                    for ordi in range(ord(val[0]), ord(val[1]) + 1):
                        out.append(form.replace(subst, chr(ordi)))
                else:
                    for num in range(int(val[0]), int(val[1])+1):
                        out.append(form.replace(subst, str(num)))
        else:
            out.append(token)

    return " ".join(out)


def tokenize(tokenstring, skip=string.whitespace,
             special=DIRECTIVES, groups=GROUPS, tgroup=TGROUP):
    """
    Parse a token string and tokenize it, yielding tokens
    """

    # This is a simplified tokenizer..
    i = -1
    while i + 1 < len(tokenstring):
        i += 1
        here = tokenstring[i]

        if here in skip:
            continue

        if here in special:
            yield here
            continue

        broken = False
        for grp in groups:
            if here == grp[0]:
                here = find_matching(grp, tokenstring[i:])
                yield here
                i += len(here)
                broken = True
                break
        if broken:
            continue

        j = i + 1
        bracket = False
        while j < len(tokenstring):
            char = tokenstring[j]
            if tgroup and char == tgroup[0]:
                bracket = True
            elif (not bracket and 
                  (char in skip or char in special)):
                break
            j += 1
            if tgroup and char == tgroup[1]:
                break

        yield tokenstring[i:j]
        i = j-1


def process(graphstring, graphs={}):
    """
    Parse a graph string construct the corresponding graph.
    """

    tokens = tokenize(preprocess(graphstring))
    directives = '(),@-=<>{}'

    G = nx.Graph()
    active = None
    parent = []
    for token in tokens:

        if token == "!":
            # Make connection to already available node (next token)
            token = next(tokens)
            if token not in G:
                raise IndexError("Token missing in graph: !{}".format(token))

        if token[0] not in directives:
            # Then it's a node (or group of nodes)

            if active is not None:
                # Extending at active node
                node = active
            elif parent:
                # Branching from parent node
                node = parent[-1]
            else:
                # Unrooted: Starting new (unconnected) graph
                node = None

            if token == '.' and node is not None:
                # Adding stub (or stub branch) to active node
                G.nodes[node]['stub'] = G.nodes[node].get('stub', 0) + 1
                LOGGER.debug("Adding stub to {}: {}", node, G.nodes[node]['stub'],
                             type='grappa')
            else:
                # Token is node or nodes
                nodes = expand_nodestring(token)
                if node is None:
                    LOGGER.debug('Unrooted nodes: {}', nodes, type='grappa')
                    G.add_nodes_from(nodes)
                else:
                    G.add_edges_from((node, n) for n in nodes)
                    LOGGER.debug("Edge: {} to {}", node, nodes, type='grappa')
                active = nodes[-1]
            continue

        # Directives:
        if token == '(':
            # Start branching
            parent.append(active)
            active = None

        elif token == ')':
            # End branch(es) - switch to active parent
            active = parent.pop()
            LOGGER.debug("End of branching: active: {}", active[-1], type='grappa')

        elif token == ',':
            # Switch to next branch
            active = None

        elif token == '@':
            # Set node as active
            active = next(tokens)
            LOGGER.debug("Setting active: {}", active, type='grappa')

        elif token == '-':
            # Remove node
            G.remove_node(next(tokens))

        elif token == '=':
            # Rename active node
            G = nx.relabel_nodes(G, {active: next(tokens)})

        elif token[0] == '<':
            # Include graph from graphs and relabel nodes according to tag
            # <tag:graphname@node>. include_graph relabels its nodes.
            B, at = include_graph(graphs, token[1:-1])
            LOGGER.debug("Including graph from {}: {}", token, B.nodes, type='grappa')
            G.add_nodes_from(B.nodes.items())
            G.add_edges_from(B.edges())
            if active is not None:
                G.add_edge(active, at)
            elif parent and parent[-1] is not None:
                G.add_edge(parent[-1], at)

        elif token[0] == '{':
            # Set attributes to active node
            LOGGER.debug("Setting attributes at active atom: {}",
                         parse_attribute_token(token), type='grappa')
            G.nodes[active].update(parse_attribute_token(token))

    return G



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Provides a Processor that identifies unexpected atoms such as PTMs and
protonations, and canonicalizes their attributes based on modifications known
in the forcefield.
"""

from collections import defaultdict
import itertools

import networkx as nx

from .processor import Processor
from ..log_helpers import StyleAdapter, get_logger
from ..utils import format_atom_string, are_all_equal

LOGGER = StyleAdapter(get_logger(__name__))


def ptm_node_matcher(node1, node2):
    """
    Returns True iff node1 and node2 should be considered equal. This means
    they are both either marked as PTM_atom, or not. If they both are PTM
    atoms, the elements need to match, and otherwise, the atomnames must
    match.
    """
    if node1.get('PTM_atom', False) == node2.get('PTM_atom', False):
        if node2.get('PTM_atom', False):
            # elements must match
            return node1['element'] == node2['element']
        else:
            # atomnames must match
            return node1['atomname'] == node2['atomname']
    else:
        return False


def find_ptm_atoms(molecule):
    """
    Finds all atoms in molecule that have the node attribute ``PTM_atom`` set
    to a value that evaluates to ``True``. ``molecule`` will be traversed
    starting at these atoms untill all marked atoms are visited such that they
    are identified per "branch", and for every branch the anchor node is known.
    The anchor node is the node(s) which are not PTM atoms and share an edge
    with the traversed branch.

    Parameters
    ----------
    molecule : networkx.Graph

    Returns
    -------
    list[tuple[set, set]]
        ``[({ptm atom indices}, {anchor indices}), ...]``. Ptm atom indices are
        connected, and are connected to the rest of molecule via anchor
        indices.
    """

    # Atomnames have already been fixed, and missing atoms have been added.
    # In addition, unrecognized atoms have been labeled with the PTM attribute.
    extra_atoms = set(n_idx for n_idx in molecule
                      if (molecule.nodes[n_idx].get('PTM_atom', False)
                          or molecule.nodes[n_idx].get('modifications')))
    ptms = []
    while extra_atoms:
        # First PTM atom we'll look at
        orig = next(iter(extra_atoms))
        anchors = set()
        # PTM atoms we've found
        atoms = set()
        # Atoms we still need to see this traversal
        to_see = set()
        # Traverse in molecule.
        while True:
            if orig in extra_atoms and orig not in atoms:
                # If this is a PTM atom, we want to see it's neighbours as
                # well.
                to_see.update(molecule[orig].keys())
                atoms.add(orig)
            elif orig not in extra_atoms:
                # Else, it's an attachment point for the this PTM
                anchors.add(orig)
            if not to_see:
                # We've traversed the interesting bit of the tree
                break
            orig = to_see.pop()
        # Although we know how far our tree spans we may still have work to do
        # for terminal nodes. There has to be a more elegant solution though.
        # for node in to_see:
        #     if node in extra_atoms:
        #         atoms.add(node)
        #     else:
        #         anchors.add(node)
        extra_atoms -= atoms
        ptms.append((atoms, anchors))
    return ptms


def identify_ptms(residue, residue_ptms, known_ptms):
    """
    Identifies all PTMs in ``known_PTMs`` nescessary to describe all PTM atoms in
    ``residue_ptms``. Will take PTMs such that all PTM atoms in ``residue``
    will be covered by applying PTMs from ``known_PTMs`` in order.
    Nodes in ``residue`` must have correct ``atomname`` attributes, and may not
    be missing. In addition, every PTM in must be anchored to a non-PTM atom.

    Parameters
    ----------
    residue : networkx.Graph
        The residues involved with these PTMs. Need not be connected.

    residue_ptms : list[tuple[set, set]]
        As returned by ``find_PTM_atoms``, but only those relevant for
        ``residue``.

    known_PTMs : collections.abc.Sequence[tuple[networkx.Graph, networkx.isomorphism.GraphMatcher]]
        The nodes in the graph must have the `PTM_atom` attribute (True or
        False). It should be True for atoms that are not part of the PTM
        itself, but describe where it is attached to the molecule.
        In addition, its nodes must have the `atomname` attribute, which will
        be used to recognize where the PTM  is anchored, or to correct the
        atomnames. Lastly, the nodes may have a `replace` attribute, which
        is a dictionary of ``{attribute_name: new_value}`` pairs. The special
        case here is if attribute_name is ``'atomname'`` and new_value is
        ``None``: in this case the node will be removed.
        Lastly, the graph (not its nodes) needs a 'name' attribute.

    Returns
    -------
    list[tuple[networkx.Graph, dict]]
        All PTMs from ``known_PTMs`` needed to describe the PTM atoms in
        ``residue`` along with a ``dict`` of node correspondences. The order of
        ``known_PTMs`` is preserved.

    Raises
    ------
    KeyError
        Not all PTM atoms in ``residue`` can be covered with ``known_PTMs``.
    """
    to_cover = set()
    cover = []
    for res_ptm in residue_ptms:
        ptm_atoms, anchors = res_ptm
        # For every node in this residue, get all modifications already known.
        residue_mods = [residue.nodes[idx].get('modifications', []) for idx in ptm_atoms]
        # Deduplicate the modifications so we only check e.g. C-ter once for
        # this residue
        used_mods = []
        for node_mods in residue_mods:
            for mod in node_mods:
                if mod not in used_mods:
                    used_mods.append(mod)
        if used_mods:
            # Store all atoms matched with already known mods (in used_mods) in
            # known_matched, so we can subtract them from ptm_atoms afterwards.
            # This may cause issues with overlapping modifications. It may be
            # better to subtract match from ptm_atoms in the loop.
            known_matched = set()
            for mod in used_mods:
                gm = nx.isomorphism.GraphMatcher(residue.subgraph(ptm_atoms), mod,
                                                 node_match=nx.isomorphism.categorical_node_match('atomname', ''))
                match = list(gm.subgraph_isomorphisms_iter())
                assert len(match) == 1
                match = match[0]
                cover.append((mod, match))
                # (That would be here)
                known_matched.update(match)
            ptm_atoms -= known_matched
            assert not ptm_atoms
        else:
            to_cover.update(ptm_atoms)
            to_cover.update(anchors)
    cover += _cover_graph(residue, to_cover, known_ptms)
    return cover


def _cover_graph(graph, to_cover, fragments):
    # BASECASE: to_cover is empty
    if not to_cover:
        return []

    # All non-PTM atoms in residue are always available for matching...
    available = set(n_idx for n_idx in graph
                    if not graph.nodes[n_idx].get('PTM_atom', False))
    # ... and add those we still need to cover
    available.update(to_cover)

    # REDUCTION: Apply one of fragments, remove those atoms from to_cover
    # COMBINATION: add the applied option to the output.
    for idx, option in enumerate(fragments):
        graphlet, matcher = option
        matches = list(matcher.subgraph_isomorphisms_iter())
        # Matches: [{graph_idxs: fragment_idxs}, {...}, ...]
        for match in matches:
            matching = set(match.keys())
            # TODO: one of the matching atoms must be an anchor. Should be
            # handled by PTMGraphMatcher already, assuming every PTM graph has
            # at least one non-ptm atom specified
            if matching <= available:
                # Continue with the remaining ptm atoms, and try just this
                # option and all smaller.
                try:
                    rest_cover = _cover_graph(graph, to_cover - matching, fragments[idx:])
                except KeyError:
                    continue
                return [(graphlet, match)] + rest_cover
    raise KeyError('Could not identify PTM')


def allowed_ptms(residue, res_ptms, known_ptms):
    """
    Finds all PTMs in ``known_ptms`` which might be relevant for ``residue``.

    Parameters
    ----------
    residue : networkx.Graph

    res_ptms : list[tuple[set, set]]
        As returned by ``find_PTM_atoms``.
        Currently not used.

    known_ptms : collections.abc.Mapping[str, networkx.Graph]

    Yields
    ------
    tuple[networkx.Graph, networkx.isomorphism.GraphMatcher]
        All graphs in known_ptms which are subgraphs of residue.
    """
    # TODO: filter by element count first
    for ptm in known_ptms.values():
        ptm_graph_matcher = nx.isomorphism.GraphMatcher(residue, ptm, node_match=ptm_node_matcher)
        if ptm_graph_matcher.subgraph_is_isomorphic():
            yield ptm, ptm_graph_matcher


def fix_ptm(molecule):
    '''
    Canonizes all PTM atoms in molecule, and labels the relevant residues with
    which PTMs were recognized. Modifies ``molecule`` such that atomnames of
    PTM atoms are corrected, and the relevant residues have been labeled with
    which PTMs were recognized.

    Parameters
    ----------
    molecule : networkx.Graph
        Must not have missing atoms, and atomnames must be correct. Atoms which
        could not be recognized must be labeled with the attribute
        PTM_atom=True.
    '''
    ptm_atoms = find_ptm_atoms(molecule)

    def key_func(ptm_atoms):
        node_idxs = ptm_atoms[-1]  # The anchors
        return sorted(molecule.nodes[idx]['resid'] for idx in node_idxs)

    ptm_atoms = sorted(ptm_atoms, key=key_func)

    resid_to_idxs = defaultdict(list)
    for n_idx in molecule:
        residx = molecule.nodes[n_idx]['resid']
        resid_to_idxs[residx].append(n_idx)
    resid_to_idxs = dict(resid_to_idxs)

    known_ptms = molecule.force_field.modifications

    for resids, res_ptms in itertools.groupby(ptm_atoms, key_func):
        # How to solve this graph covering problem
        # Filter known_ptms, such that
        #   element_count(known_ptm) <= element_count(found)
        # Filter known_ptms, such that known_ptm <= found (subgraph of).
        #   Note that this does mean that the PTMs in the PDB *must* be
        #   complete. So no missing atoms.
        # Find all the exactly covering combinations.
        # Pick the best solution, such that the maximum size of the applied
        # PTMs is maximal. (3, 2) > (3, 1, 1) > (2, 2, 1)
            # Numbers are sizes of applied PTMs

        # The last two steps are combined by recursively trying the largest
        # option in identify_ptms

        res_ptms = list(res_ptms)
        n_idxs = set()
        for resid in resids:
            n_idxs.update(resid_to_idxs[resid])
        # TODO: Maybe use graph_utils.make_residue_graph? Or rewrite that
        #       function?
        residue = molecule.subgraph(n_idxs)
        options = allowed_ptms(residue, res_ptms, known_ptms)
        options = sorted(options,
                         key=lambda opt: len([n for n in opt[0]
                                              if opt[0].nodes[n].get('PTM_atom', False)]),
                         reverse=True)
        try:
            identified = identify_ptms(residue, res_ptms, options)
        except KeyError:
            LOGGER.warning('Could not identify the modifications for'
                           ' residues {}, involving atoms {}',
                           ['{resname}{resid}'.format(**molecule.nodes[resid_to_idxs[resid][0]])
                            for resid in sorted(set(resids))],
                           ['{atomid}-{atomname}'.format(**molecule.nodes[idx])
                            for idxs in res_ptms for idx in idxs[0]],
                           type='unknown-input')
            for idxs in res_ptms:
                for idx in idxs[0]:
                    molecule.remove_node(idx)
            continue

        # Why this mess? There can be multiple PTMs for a single (set of)
        # residue(s); and a single PTM can span multiple residues.
        LOGGER.info("Identified the modifications {} on residues {}",
                    [out[0].graph['name'] for out in identified],
                    ['{resname}{resid}'.format(**molecule.nodes[resid_to_idxs[resid][0]])
                     for resid in resids])
        for ptm, match in identified:
            ptm.match = match
            for mol_idx, ptm_idx in match.items():
                ptm_node = ptm.nodes[ptm_idx]
                mol_node = molecule.nodes[mol_idx]
                # Names of PTM atoms still need to be corrected, and for some
                # non PTM atoms attributes need to change.
                if ptm_node['PTM_atom']:
                    mol_node['graph'] = molecule.subgraph([mol_idx]).copy()
                    for attr in ptm_node:
                        # FIXME: This probably transfers too many attributes.
                        if attr not in ('PTM_atom', 'replace'):
                            mol_node[attr] = ptm_node[attr]
                if 'replace' in ptm_node:
                    to_replace = ptm_node['replace']
                    for attr_name, val in to_replace.items():
                        if attr_name == 'atomname':
                            mol_node['_old_atomname'] = mol_node['atomname']
                        # We can't remove nodes which get their atomname set to
                        # None here, since mapping would then break due to
                        # missing atoms. Instead, the mapping processor should
                        # make sure superfluous atoms do not get constructed in
                        # the output resolution.
                        if mol_node.get(attr_name) != val:
                            fmt = 'Changing attribute {} from {} to {} for atom {}'
                            LOGGER.debug(fmt, attr_name, mol_node.get(attr_name),
                                         val, format_atom_string(mol_node),
                                         type='change-atom')
                            mol_node[attr_name] = val
            for n_idx in n_idxs:
                node = molecule.nodes[n_idx]
                if not ('modification' in node and ptm in node.get('modifications', [])):
                    # These nodes already had the modification annotated.
                    # Also note that 'modification' != 'modifications'. Yes,
                    # this is an issue. No, I'm not fixing that.
                    node['modifications'] = node.get('modifications', [])
                    node['modifications'].append(ptm)


class CanonicalizeModifications(Processor):
    def run_molecule(self, molecule):
        fix_ptm(molecule)
        return molecule



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the :class:`~vermouth.processors.set_molecule_meta.SetMoleculeMeta` processor.
"""

import copy

import pytest

from vermouth import System
from vermouth.molecule import Molecule
from vermouth.processors.set_molecule_meta import SetMoleculeMeta
from vermouth.utils import are_different


@pytest.mark.parametrize('prior, meta, expected', (
    (
        {'foo': 'bar', 'toto': [3, 4, None]},
        {'new key': 9.0},
        {'foo': 'bar', 'toto': [3, 4, None], 'new key': 9.0},
    ),
    (
        {'foo': 'bar', 'toto': [3, 4, None]},
        {'foo': 'other', 'new': 90},
        {'foo': 'other', 'toto': [3, 4, None], 'new': 90},
    ),
    (
        {},
        {'foo': 'bar', 'toto': 'tata'},
        {'foo': 'bar', 'toto': 'tata'},
    ),
    (
        {'foo': 'bar', 'toto': 'tata'},
        {},
        {'foo': 'bar', 'toto': 'tata'},
    ),
))
def test_set_molecule_meta_molecule(prior, meta, expected):
    """
    Test for :func:`SetMoleculeMeta.run_molecule` .
    """
    molecule = Molecule()
    molecule.meta = copy.copy(prior)
    processor = SetMoleculeMeta(**meta)
    processor.run_molecule(molecule)
    assert not are_different(molecule.meta, expected)


@pytest.mark.parametrize('priors, meta, expected', (
    (
        [{}, {'foo': 'bar'}, {'toto': {'a': 0, 'b': None}}],
        {'new': 42, 'foo': None},
        [
            {'new': 42, 'foo': None},
            {'new': 42, 'foo': None},
            {'new': 42, 'foo': None, 'toto': {'a': 0, 'b': None}},
        ],
    ),
))
def test_set_molecule_meta_system(priors, meta, expected):
    """
    Test for :func:`SetMoleculeMeta.run_system` .
    """
    system = System()
    for prior in priors:
        molecule = Molecule()
        molecule.meta = copy.copy(prior)
        system.add_molecule(molecule)
    processor = SetMoleculeMeta(**meta)
    processor.run_system(system)
    for molecule, expectation in zip(system.molecules, expected):
        assert not are_different(molecule.meta, expectation)



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the geometry module.
"""

import itertools
import numpy as np
import pytest
from vermouth import geometry


def _generate_test_angles(n_angles):
    """
    Gererate a series of coordinates of 3 points at a different angles.

    Generate 'n_angles' structures with angles between 0 and pi with regular
    angle spacing.

    Parameters
    ----------
    n_angles: int
        Number of structures to generate.

    Yields
    ------
    coordinates: np.ndrarray
        The coordinates of the points. Each row corresponds to a pointm and
        each column corresponds to a dimension.
    angle: float
        The angle between the 3 points.
    """
    # The shift is arbitrary. Its purpose is to avoid having the angle centered
    # at the origin as it may be a special case.
    shift = np.array([1.1, -7.2, 9.1])
    coordinates = np.array([
        [2, 0, 0],
        [0, 0, 0],
        [0, 0, 0],  # Will be redifined
    ]) + shift
    # The distance of the 3rd point to the second one, also the radius in polar
    # coordinates. The value is totally arbitrary, and it should not change the
    # angle; but avoid setting the radius to 1 as it may hide normalization
    # errors.
    radius = 4.2
    for angle in np.linspace(0, np.pi, num=n_angles):
        coordinates[-1, 0] = radius * np.cos(angle) + shift[0]
        coordinates[-1, 1] = radius * np.sin(angle) + shift[1]
        yield coordinates.copy(), angle


def _generate_test_dihedrals(n_angles):
    """
    Generate a series of coordinates for 4 points at different torsion angles.

    Generate 'n_angles' structures with torsion angles between -pi and +pi
    with regular angle spacing.

    Parameters
    ----------
    n_angles: int
        Number of structures to generate.

    Yields
    ------
    coordinates: np.ndrarray
        The coordinates of the points. Each row corresponds to a pointm and
        each column corresponds to a dimension.
    angle: float
        The angle around the ais between the middle points.
    """
    # The shift is arbitrary. Its purpose is to avoid having the angle centered
    # at the origin as it may be a special case.
    shift = np.array([1.1, -7.2, 9.1])
    coordinates = np.array([
        [2, 0, 0],
        [0, 0, 0],
        [0, 0, 5],
        [0, 0, 5],  # Will be redifined
    ]) + shift
    # The distance of the 3rd point to the second one, also the radius in polar
    # coordinates. The value is totally arbitrary, and it should not change the
    # angle; but avoid setting the radius to 1 as it may hide normalization
    # errors.
    radius = 4.2
    for angle in np.linspace(-np.pi, +np.pi, num=n_angles):
        coordinates[-1, 0] = radius * np.cos(angle) + shift[0]
        coordinates[-1, 1] = radius * np.sin(angle) + shift[1]
        yield coordinates.copy(), angle


@pytest.mark.parametrize(
    'points, angle',
    itertools.chain(
        _generate_test_angles(10),
        ((np.array([[0,  3, 0], [0, 0, 0], [0, 6, 0]]), 0), ),  # pylint: disable=bad-whitespace
        ((np.array([[0, -9, 0], [0, 0, 0], [0, 2, 0]]), np.pi), ),
    )
)
def test_angle(points, angle):
    vectorBA = points[0, :] - points[1, :]
    vectorBC = points[2, :] - points[1, :]
    assert np.allclose(geometry.angle(vectorBA, vectorBC), angle)


@pytest.mark.parametrize(
    'points, angle',
    itertools.chain(
        _generate_test_dihedrals(10),
        ((np.array([[0,  3, 0], [0, 0, 0], [4, 0, 0], [7, 6, 0]]), 0), ),  # pylint: disable=bad-whitespace
        ((np.array([[0, -9, 0], [0, 0, 0], [4, 0, 0], [5, 6, 0]]), np.pi), ),
    )
)
def test_dihedral(points, angle):
    calc_angle = geometry.dihedral(points)
    # +pi and -pi are the same angle; we normalize them to pi
    if np.allclose(calc_angle, -np.pi):
        calc_angle *= -1
    if np.allclose(angle, -np.pi):
        angle *= -1
    assert np.allclose(calc_angle, angle)


@pytest.mark.parametrize(
    'points, angle',
    itertools.chain(
        _generate_test_dihedrals(10),
        ((np.array([[0,  3, 0], [0, 0, 0], [4, 0, 0], [7, 6, 0]]), 0), ),  # pylint: disable=bad-whitespace
        ((np.array([[0, -9, 0], [0, 0, 0], [4, 0, 0], [5, 6, 0]]), np.pi), ),
    )
)
def test_dihedral_phase(points, angle):
    angle_phase = angle + np.pi
    if angle_phase > np.pi:
        angle_phase -= 2 * np.pi
    if angle_phase < -np.pi:
        angle_phase += 2 * np.pi
    calc_angle = geometry.dihedral_phase(points)
    # +pi and -pi are the same angle; we normalize them to pi
    if np.allclose(calc_angle, -np.pi):
        calc_angle *= -1
    if np.allclose(angle_phase, -np.pi):
        angle_phase *= -1
    assert np.allclose(calc_angle, angle_phase)


def test_distance_matrix():
    # This array of coordinates was generated using:
    #    coordinates = (
    #        np.random.uniform(low=-2, high=2, size=(15, 3))
    #        .astype(np.float32)
    #        .round(2)
    #    )
    #    coordinates.tolist()
    coordinates = np.array([
        [-0.4099999964237213, 0.5699999928474426, -1.2000000476837158],
        [-0.05000000074505806, 1.6799999475479126, 1.0800000429153442],
        [1.649999976158142, -1.2699999809265137, -0.18000000715255737],
        [-0.4300000071525574, -0.5, -1.25],
        [1.75, -0.28999999165534973, -1.75],
        [1.399999976158142, 1.0399999618530273, -0.029999999329447746],
        [0.25999999046325684, -0.9399999976158142, -0.8899999856948853],
        [1.159999966621399, -0.10999999940395355, 0.07999999821186066],
        [0.009999999776482582, 0.5199999809265137, -0.8600000143051147],
        [-1.909999966621399, 1.5, -0.27000001072883606],
        [-1.5800000429153442, 0.5299999713897705, -0.6499999761581421],
        [1.9900000095367432, -1.7799999713897705, 1.7699999809265137],
        [1.6100000143051147, -0.03999999910593033, -0.8899999856948853],
        [0.10000000149011612, 1.1200000047683716, 0.17000000178813934],
        [0.41999998688697815, -0.5699999928474426, 0.33000001311302185],
    ])

    # The reference was built using MDAnalysis:
    #     mad.distance_array(coordinates[:6], coordinates[6:15]).tolist()
    reference = np.array([
        [1.6808033650982364, 2.1367498758730443, 0.5426785539106965,
         1.9949436426807041, 1.293445112450694, 4.483681543049364,
         2.1327447054976916, 1.5618898861478303, 2.0807210809989],
        [3.2926280559049035, 2.3807981156038913, 2.2611501818462068,
         2.3053199558207305, 2.5799806759724233, 4.075450898096163,
         3.0975635086704183, 1.0789809940089903, 2.417829605013437],
        [1.5953369295377542, 1.285807111097364, 2.5211306642537474,
         4.511607195619645, 3.727438788554226, 2.044064626898197,
         1.4207744414223977, 2.8700347089662435, 1.504327098906351],
        [0.8940357962136768, 2.109288939442685, 1.1773274598997576,
         2.674097996304839, 1.6563212819432889, 4.076174699496656,
         2.1219801748218843, 2.21849046119744, 1.7954944096661205],
        [1.8390758608686804, 1.931165501366316, 2.115608639381968,
         4.334754756955464, 3.601569024030028, 3.8298955467113966,
         0.9064767071466125, 2.8977576955983215, 2.4846930992457765],
        [2.4412292124092203, 1.1799152298407856, 1.7004117182714393,
         3.3504178337743813, 3.086243689319313, 3.397131067712101,
         1.3964597652032842, 1.31772526881778, 1.9188798060046086]
    ])

    matrix = geometry.distance_matrix(coordinates[:6], coordinates[6:15])
    assert matrix.shape == (6, 9)
    assert np.allclose(matrix, reference)
